from .recursiveagent import ETClient
from .factsrag2 import FactRAGFileCache, FactBot, clean_fact_citations
from .utils import UtilityHelper
from . import Client, Statement, Forecast

from phasellm.llms import ChatBot, OpenAIGPTWrapper, ChatPrompt

from datetime import datetime

start_system_prompt = """Today's date is {the_date}. You are a researcher helping with economics and politics research. We will give you a few facts and we need you to fill in a blank to the best of your knowledge, based on all the information provided to you. All your answers should be absed on these facts ONLY.

For example, suppose we ask, 'Who is the President of the USA?' and have the following facts...

F1: The President of the USA is Joe Biden.
F2: The Vice President of the USA is Kamala Harris.

... your answers hould be something like this:

The President of th USA is Joe Biden [F1].

We will give you a list of facts for every question. You can reference those facts, or you can also reference earlier facts from the conversatio chain. YOU CANNOT USE OTHER INFORMATION."""

start_user_prompt = """Here is the research:
{content}
{additional_facts}
------------

Given the above, we need you to do your best to fill in the following blank...
{fill_in_the_blank}

PLEASE DO THE FOLLOWING:
- Provide any further justification ONLY BASED ON THE FACTS AND SOURCES PROVIDED ABOVE.
- Explain your forecast and how the facts, insights, etc. support it. Do not simply state a number.
- Do not provide a range; provide ONE number.
- End your forecast with the filled-in statement: {fill_in_the_blank_2}

We realize you are being asked to provide a speculative forecast. We are using this to better understand the world and finance, so please fill in the blank. We will not use this for any active decision-making, but more to learn about the capabilities of AI.
"""

extend_user_prompt = """Here is the research:
{content}
{additional_facts}
---------------------

In addition to the new content above, we want to UPDATE the forecast from before. Here is the earlier forecast...
---------------------
FORECAST: {earlier_forecast_value}

JUSTIFICATION:
{earlier_forecast}
---------------------

Given the above, we need you to do your best to fill in the following blank...
{fill_in_the_blank}

PLEASE DO THE FOLLOWING:
- Provide any further justification ONLY BASED ON THE FACTS AND SOURCES PROVIDED ABOVE.
- Explain your forecast and how the facts, insights, etc. support it. Do not simply state a number.
- Do not provide a range; provide ONE number.
- End your forecast with the filled-in statement: {fill_in_the_blank_2}

We realize you are being asked to provide a speculative forecast. We are using this to better understand the world and finance, so please fill in the blank. We will not use this for any active decision-making, but more to learn about the capabilities of AI."""


class FactsRAGForecastingAgent(object):

    # TODO: document / clean up
    def __init__(
        self,
        client: ETClient,
        chatbot: ChatBot,
        factbase: FactRAGFileCache,
    ):

        self.client = client
        self.chatbot = chatbot
        self.factbase = factbase

    # TODO / NOTE: this allows us to continue chatting with the forecasting agent, since we can obtain the chatbot later. Given that some folks are interested in asking for clarifications, this could be an interesting opportunity.
    def setChatBot(self, chatbot):
        self.chatbot = chatbot

    # TODO: standardize -- camel case or snake case? Or something else?
    def getChatBot(self):
        return self.chatbot

    # TODO: we can do much better at disaggregating all these functions. Currently just want this to work.
    def create_forecast(
        self,
        statement: Statement,
        openai_api_key,
        et_api_key,
        facts=None,
        prediction_agent="Test Agent",
    ):

        # factbot = FactBot(self.factbase, openai_api_key)
        query1 = self.factbase.query_to_fact_content(
            statement.fill_in_the_blank, n_results=25, skip_separator=True
        )
        query2 = self.factbase.query_to_fact_content(
            statement.description, n_results=25, skip_separator=True
        )

        if len(query1) == 0 and len(query2) == 0:
            print("No new content added to the forecast.")
            return None

        facts_to_use = (
            """--- START FACTS ---------------------------\n"""
            + query1.strip()
            + "\n"
            + query2.strip()
            + """--- END FACTS ---------------------------\n"""
        )

        chatbot_messages = [
            {"role": "system", "content": start_system_prompt},
            {"role": "user", "content": start_user_prompt},
        ]

        chatbot = self.chatbot

        prompt_template = ChatPrompt(chatbot_messages)

        the_date = datetime.now().strftime("%Y-%m-%dT%H:%M:%S")

        additional_facts = ""
        if facts is not None:
            additional_facts = "Some additional facts for consideration are below...\n"
            afctr = 1
            for f in facts:
                additional_facts += f"AF{afctr}: {f}\n"
                afctr += 1
            additional_facts += "---------------------\n\n"

        chatbot.messages = prompt_template.fill(
            statement_title=statement.title,
            statement_description=statement.description,
            statement_fill_in_the_blank=statement.fill_in_the_blank,
            fill_in_the_blank_2=statement.fill_in_the_blank,
            content=facts_to_use,
            the_date=the_date,
            additional_facts=additional_facts,
        )

        assistant_analysis = chatbot.resend()
        full_content = clean_fact_citations(self.factbase, assistant_analysis)

        print("\n\n\n")
        print(assistant_analysis)

        uh = UtilityHelper(openai_api_key)
        prediction = uh.extract_prediction(
            assistant_analysis, statement.fill_in_the_blank
        )

        client = Client(et_api_key)

        # full_content = content + "\n\n-----------------\n\n" + assistant_analysis

        response = client.create_forecast(
            statement.id,
            "Prediction",
            full_content,
            prediction,
            prediction_agent,
            {
                # "full_response_from_llm_before_source_cleanup": content,
                "full_response_from_llm": assistant_analysis,
                "extracted_value": prediction,
            },
        )

        return response

    def extend_forecast(
        self,
        forecast: Forecast,
        openai_api_key,
        et_api_key,
        facts=None,
        prediction_agent="Test Agent",
    ):

        # Note: we only update the forecast with data/info we added since the last forecast.

        query1 = self.factbase.query_to_fact_content(
            forecast.statement.fill_in_the_blank,
            n_results=25,
            skip_separator=True,
            since_date=forecast.created_at,
        )
        query2 = self.factbase.query_to_fact_content(
            forecast.statement.description,
            n_results=25,
            skip_separator=True,
            since_date=forecast.created_at,
        )

        if len(query1) == 0 and len(query2) == 0:
            print("No new content added to the forecast.")
            return None

        facts_to_use = (
            """--- START FACTS ---------------------------\n"""
            + query1.strip()
            + "\n"
            + query2.strip()
            + """--- END FACTS ---------------------------\n"""
        )

        chatbot_messages = [
            {"role": "system", "content": start_system_prompt},
            {"role": "user", "content": extend_user_prompt},
        ]

        chatbot = self.chatbot

        prompt_template = ChatPrompt(chatbot_messages)

        the_date = datetime.now().strftime("%Y-%m-%dT%H:%M:%S")

        additional_facts = ""
        if facts is not None:
            additional_facts = "Some additional facts for consideration are below...\n"
            afctr = 1
            for f in facts:
                additional_facts += f"AF{afctr}: {f}\n"
                afctr += 1
            additional_facts += "---------------------\n\n"

        chatbot.messages = prompt_template.fill(
            statement_title=forecast.statement.title,
            statement_description=forecast.statement.description,
            statement_fill_in_the_blank=forecast.statement.fill_in_the_blank,
            fill_in_the_blank_2=forecast.statement.fill_in_the_blank,
            content=facts_to_use,
            the_date=the_date,
            additional_facts=additional_facts,
            earlier_forecast_value=str(forecast.value),
            earlier_forecast=forecast.justification,
        )

        assistant_analysis = chatbot.resend()

        # print("\n\n\n")
        # print(assistant_analysis)

        full_content = clean_fact_citations(self.factbase, assistant_analysis)
        print(full_content)

        uh = UtilityHelper(openai_api_key)
        prediction = uh.extract_prediction(
            assistant_analysis, forecast.statement.fill_in_the_blank
        )

        client = Client(et_api_key)

        response = client.create_forecast(
            forecast.statement.id,
            "Prediction",
            full_content,
            prediction,
            prediction_agent,
            {
                "full_response_from_llm": assistant_analysis,
                "extracted_value": prediction,
            },
            forecast.id,
        )

        return response



import requests
import feedparser
import time
import random

from .crawlers import crawlerPlaywright, _get_text_bs4

from playwright.sync_api import sync_playwright

from news_search_client import NewsSearchClient
from azure.core.credentials import AzureKeyCredential


def force_empty_content(rss_url: str, content, cache_function) -> None:
    """
    Force the crawler to visit every URL in the RSS feed and save it as a blank content file. We do this because some RSS feeds have a lot of old URLs we do not need to crawl, and only want to crawl the delta over some period.

    Args:
        rss_url (str): The URL of the RSS feed.
        content: the content string to save.
        cache_function: the specific function to call the rss_url and content to save.
    """

    agent = RSSAgent(rss_url)
    all_urls = agent.get_news_as_list()
    for u in all_urls:
        cache_function(u, content)


class RSSAgent:

    def __init__(self, rss_url, crawler=None) -> None:
        """
        A simple wrapper for an RSS feed, so we can query it for URLs.

        Args:
            rss_url (str): The URL of the RSS feed.
            crawler (Crawler, optional): The crawler to use. Defaults to None, in which case we will use crawlerPlaywright in headless mode.
        """
        self.rss_url = rss_url
        if crawler is None:
            self.crawler = crawlerPlaywright()
        else:
            self.crawler = crawler

    def get_news_as_list(self) -> list:
        """
        Query the RSS feed for news articles, and return them as a list of dictionaries.

        Returns:
            list: A list of URLs.
        """
        urls = []
        feed = feedparser.parse(self.rss_url)
        for entry in feed.entries:
            urls.append(entry.link)
        return urls


class NewsBingAgent:

    def __init__(self, api_key: str, endpoint: str):
        """
        Creates a new Bing News API agent. To learn more, see: https://github.com/microsoft/bing-search-sdk-for-python/

        Args:
            api_key (str): The Bing News API key.
            endpoint (str): The Bing News API endpoint.
        """
        self.api_key = api_key
        self.endpoint = endpoint

    def get_news_as_list(self, query: str, market: str = "en-us") -> list:
        """
        Gets a list of URLS from the Bing News API. For more information on markets, see: https://learn.microsoft.com/en-us/bing/search-apis/bing-web-search/reference/market-codes

        Args:
            query (str): The query to search for.
            market (str, optional): The market to search in. Defaults to "en-us". (US English

        Returns:
            list: A list of URLs.
        """

        client = NewsSearchClient(
            endpoint=self.endpoint, credential=AzureKeyCredential(self.api_key)
        )

        urls = []

        try:
            news_result = client.news.search(query=query, market=market, count=10)
            for n in news_result.value:
                urls.append(n.url)

        except Exception as err:
            print("Encountered exception. {}".format(err))

        return urls


class NewsAPIAgent:

    def __init__(self, api_key, top_headlines=False, crawler=None) -> None:
        """
        A simple wrapper for the News API, so we can query it for URLs.

        Args:
            api_key (str): The News API key.
            top_headlines (bool, optional): Whether to get top headlines. Defaults to False.
            crawler (Crawler, optional): The crawler to use. Defaults to None, in which case we will use crawlerPlaywright in headless mode.
        """
        self.api_key = api_key
        self.top_headlines = top_headlines
        if crawler is None:
            self.crawler = crawlerPlaywright()
        else:
            self.crawler = crawler

    def get_news_as_list(self, query: str) -> list:
        """
        Query the News API for news articles, and return them as a list of dictionaries.

        Args:
            query (str): The query to search for.

        Returns:
            list: A list of dictionaries, where each dictionary represents a news article.
        """
        url = f"https://newsapi.org/v2/everything?q={query}&apiKey={self.api_key}"
        if self.top_headlines:
            url = (
                f"https://newsapi.org/v2/top-headlines?q={query}&apiKey={self.api_key}"
            )
        response = requests.get(url)
        return response.json()


class FinancialTimesAgent:

    # The RSS feed URLs for the Financial Times.
    ft_rss_feed_urls = [
        "https://www.ft.com/rss/home",
        "https://www.ft.com/world?format=rss",
        "https://www.ft.com/global-economy?format=rss",
        "https://www.ft.com/companies?format=rss",
        "https://www.ft.com/opinion?format=rss",
    ]

    ft_login_url = "https://ft.com/login"
    ft_main_url = "https://ft.com/"

    def __init__(self, user_email, user_password) -> None:
        """
        This is a POC agent that uses Playwright to crawl the Financial Times articles you are interested in. Note that you *NEED* to be a subscriber to the FT to make this work, and thus need to provide your FT user name and password.

        Args:
            user_email (str): Your FT email.
            user_password (str): Your FT password.
        """
        self.user_email = user_email
        self.user_password = user_password

    def get_news(self, urls: list[str] = None) -> list:
        """
        Get the news from the Financial Times as a list of tuples, where each tuple contains the URL and the extracted text content.

        Args:
            urls: a list of URLs to get content for.

        Returns:
            A list of lists -- urls, html, and text content
        """

        if urls is None:
            urls = set()
            for rss_url in self.ft_rss_feed_urls:
                agent = RSSAgent(rss_url)
                rss_url_list = agent.get_news_as_list()
                for r in rss_url_list:
                    urls.add(r)
            urls = list(urls)

        html_content_array = []
        text_content_array = []

        with sync_playwright() as playwright:

            browser = playwright.firefox.launch(headless=False)
            page = browser.new_page()

            # Navigate to the webpage
            page.goto(self.ft_main_url)

            print("Accepting Cookies")
            page.frame_locator('*[title="SP Consent Message"]').get_by_text(
                "Accept Cookies"
            ).click()

            time.sleep(2)

            page.goto(self.ft_login_url)

            time.sleep(2)

            print("Entering user name + hitting enter")

            page.locator("#enter-email").fill(self.user_email)
            page.keyboard.press("Enter")

            time.sleep(5)

            page.locator("#enter-password").fill(self.user_password)
            page.keyboard.press("Enter")

            time.sleep(5)

            url_ctr = 1
            for url in urls:
                print(f"Getting content for URL {url_ctr} of {len(urls)}")

                html_content = ""
                text_content = ""

                try:
                    page.goto(url)
                    html_content = page.content()
                    text_content = _get_text_bs4(html_content)

                    print(url)
                    print(text_content)

                except:
                    print(url)
                    print(f"Error getting content for URL {url_ctr} of {len(urls)}")

                html_content_array.append(html_content)
                text_content_array.append(text_content)

                url_ctr += 1

                time.sleep(2 + random.randint(0, 5))

            # Close the browser
            browser.close()

        return urls, html_content_array, text_content_array



"""
This is an experimental approach to tracking information regardless of source type. It will also power more than recent updates. Here's how it works...

1. All "Content Sources" (a new class type that obtains content) will send content directly to the Facts DB.
2. The "Facts DB" will then extract all relevant facts for a prediction or research theme. It will keep cache the original content, will track the sources, and will also input all the facts into a RAG database.
3. We can then query the DB for relevant facts on an ad hoc basis, rather than only for new content.

"""

import os
import json
import hashlib
import re

# New libraries for FAISS, etc.
import numpy as np
import faiss
import pickle
import warnings

from openai import OpenAI
import tiktoken

# Using JSONEncoder to be consistent with the Emerging Trajectories website and platform.
from django.core.serializers.json import DjangoJSONEncoder

from phasellm.llms import OpenAIGPTWrapper, ChatBot, ChatPrompt
from phasellm.agents import WebpageAgent, WebSearchAgent

from datetime import datetime, timedelta

from . import Client
from .crawlers import crawlerPlaywright
from .prompts import *
from .news import NewsAPIAgent, RSSAgent, FinancialTimesAgent, NewsBingAgent
from .chunkers import *

# Number of search results to return from web searche (default value).
_DEFAULT_NUM_SEARCH_RESULTS = 10

facts_base_system_prompt = """You are a researcher tasked with helping forecast economic and social trends. The title of our research project is: {statement_title}.

The project description is as follows...
{statement_description}

We will provide you with content from reports and web pages that is meant to help with the above. We will ask you to review these documents, create a set of bullet points to inform your thinking. Rather than using bullet points, please list each as F1, F2, F3, etc... So that we can reference it.

The content we provided you contains source numbers in the format 'SOURCE: #'. When you extract facts, please include the citation in square brackets, with the #, like [#], but replace "#" with the actual Source # from the crawled content we are providing you.

For example, if you are referring to a fact that came under --- SOURCE: 3 ---, you would write something like: "Data is already trending to hotter temperatures [3]." Do not include the "#" in the brackets, just the number.

Thus, a bullet point would look like this:
F1: (information) [1]
F2: (information) [1]
F3: (information) [2]

... and so on, where F1, F2, F3, etc. are facts, and [1], [2] are the source documents you are extracting the facts from.
"""

facts_base_user_prompt = """Today's date is {the_date}. We will now provide you with all the content we've managed to collect. 

----------------------
{scraped_content}
----------------------

Please think step-by-step by (a) extracting critical bullet points from the above, and (b) share any insights you might have based on the facts.

The content we provided you contains source numbers in the format 'SOURCE: #'. When you extract facts, please include the citation in square brackets, with the #, like [#], but replace "#" with the actual Source # from the crawled content we are providing you.

For example, if you are referring to a fact that came under --- SOURCE: 3 ---, you would write something like: "Data is already trending to hotter temperatures [3]." Do not include the "#" in the brackets, just the actual number.

DO NOT PROVIDE A FORECAST, BUT SIMPLY STATE AND SHARE THE FACTS AND INSIGHTS YOU HAVE GATHERED.
"""


def uri_to_local(uri: str) -> str:
    """
    Convert a URI to a local file name. In this case, we typically will use an MD5 sum.

    Args:
        uri (str): The URI to convert.

    Returns:
        str: The MD5 sum of the URI.
    """
    uri_md5 = hashlib.md5(uri.encode("utf-8")).hexdigest()
    return uri_md5


class VectorDBDict:
    """
    This is a Python dictionary that gets converted into a FAISS index, and gets pickled to disk. This also comes with some "DB-like" features:
    (a) a unique, autoincrementing index
    (b) no ability to delete items (for now)
    (c) additional meta data you can save
    (d) TODO, maybe: locks to prevent concurrent writes

    Pickling presents some security issues, so please be careful if you are sharing pickled data files or being provided such files from others.
    """

    # Right now, we are only working with OpenAI's 'text-embedding-3-small' model.
    VECTOR_SIZE = 1536
    MAX_BATCH_SIZE = 100

    def __init__(
        self,
        db_file_path: str,
        openai_api_key: str,
        error_out_on_conflict: bool = False,
    ) -> None:
        """
        Initialize the database.

        Args:
            db_file_path (str): The path to the database file. Will be created if it does not exist.
            openai_api_key (str): The OpenAI API key.
            error_out_on_conflict (bool): If True, we will error out if the database tries to write to a file that doesn't align with the SHA hash of the DB file when it was loaded. Basically like a lock without actually being a lock. When it's set to False, it will simply print an error.
        """

        self.db_file_path = db_file_path
        self.error_out_on_conflict = error_out_on_conflict
        self.openai_api_key = openai_api_key
        self.openai_client = OpenAI(api_key=self.openai_api_key)
        self.encoding = tiktoken.encoding_for_model("text-embedding-3-small")

        if not os.path.exists(db_file_path):
            dictionary_to_write = {
                "embeddings": faiss.IndexFlatL2(self.VECTOR_SIZE),
                "original_texts": [],
                "metadata": [],
            }
            pickle.dump(
                dictionary_to_write, open(db_file_path, "wb", pickle.HIGHEST_PROTOCOL)
            )

        self.db = pickle.load(open(db_file_path, "rb"))
        self.db_hash = self.get_file_sha256(db_file_path)

    def add_vectors(
        self, vectors: np.array, texts: list, metadata: list = None
    ) -> list:
        """
        Add vectors to the database.

        Args:
            vectors (np.array): The vectors to add.
            texts (list): The text for each vector.
            metadata (list, optional): The metadata for each vector.

        Returns:
            list: The IDs of the vectors.
        """

        start_index = len(self.db["original_texts"])

        if metadata is None:
            metadata = [{} for i in range(0, len(texts))]

        self.db["embeddings"].add(np.array(vectors))
        self.db["original_texts"] += texts
        self.db["metadata"] += metadata

        end_index = len(self.db["original_texts"])

        return list(range(start_index, end_index))

    def add_vector(self, vector: np.array, text: str, metadata: dict = None) -> int:
        """
        Adds a vector to the database.

        Args:
            vector (np.array): The vector to add.
            text (str): The text for the vector.
            metadata (dict): The metadata for the vector.

        Returns:
            int: The ID of the vector.
        """
        return self.add_vectors([vector], [text], [metadata])[0]

    def shorten_text(self, text, max_token_length: int = 8000) -> str:
        """
        Shortens text to a maximum token length. This is useful for OpenAI API calls, which have a token limit.

        Args:
            text (str): The text to shorten.
            max_token_length (int): The maximum token length.

        Returns:
            str: The shortened text; should ONLY be used for encoding.
        """

        new_text = text[:]
        while len(self.encoding.encode(new_text)) > max_token_length:
            if len(new_text) > 250:
                new_text = new_text[:-250] + "..."

        return new_text

    def add_texts(self, texts: list, metadata: list = None) -> list:
        """
        Adds text to the database. Calls an embedding function and then adds via add_vectors().

        Args:
            texts (list): The texts to add.
            metadata (list): Metadata to add.

        Returns:
            list: The IDs of the texts.
        """

        last_index = 0
        embeddings = []
        while last_index < len(texts):

            new_modifier = min(self.MAX_BATCH_SIZE, len(texts) - last_index)

            # In this approach, we keep the original text even though we encode the new text. We might want to revisit this to keep the new text.
            # TODO See above.
            text_subset = []
            text_subset_prechecked = texts[last_index : (last_index + new_modifier)]
            for t in text_subset_prechecked:
                t_new = self.shorten_text(t)
                text_subset.append(t_new)

            response = self.openai_client.embeddings.create(
                input=text_subset, model="text-embedding-3-small"
            )
            embeddings += list(map(lambda x: x.embedding, response.data))

            last_index += new_modifier

        return self.add_vectors(embeddings, texts, metadata)

    def add_text(self, text: str, metadata: dict = None) -> int:
        """
        Adds text to the database. Calls an embedding function and then adds via add_vectors().

        Args:
            text (str): The text to add.
            metadata (dict): Metadata to add.

        Returns:
            str: The ID of the text.
        """
        return self.add_texts([text], [metadata])[0]

    def get_file_sha256(self, file_path: str) -> str:
        """
        Get the SHA256 hash of a file. We use this to warn the user if/when the database is being saved and potetially conflicts with the underlying file. Written by GitHub copilot! 🙌

        Args:
            file_path (str): The path to the file.

        Returns:
            str: The SHA256 hash of the file.
        """
        sha256_hash = hashlib.sha256()
        with open(file_path, "rb") as f:
            for byte_block in iter(lambda: f.read(4096), b""):
                sha256_hash.update(byte_block)
        return sha256_hash.hexdigest()

    def save(self):
        """
        Save the database to disk.
        """
        file_sha = self.get_file_sha256(self.db_file_path)
        if file_sha != self.db_hash:
            if self.error_out_on_conflict:
                raise ValueError(
                    "The database file has been modified since it was loaded. Please reload the database and try again."
                )
            else:
                warnings.warn(
                    "WARNING: The database file has been modified since it was loaded. Please reload the database and try again."
                )

        pickle.dump(self.db, open(self.db_file_path, "wb", pickle.HIGHEST_PROTOCOL))
        self.db_hash = self.get_file_sha256(self.db_file_path)

    def count(self) -> int:
        """
        Returns the size of the DB.
        """
        return len(self.db["original_texts"])

    def get(self, index: int) -> dict:
        """
        Returns the text and metadata for a given index.

        Args:
            index (int): The index to get.

        Returns:
            dict: The text and metadata for the index.
        """

        text = self.db["original_texts"][index]
        metadata = self.db["metadata"][index]

        return text, metadata

    def query(self, text: str, n: int = 10) -> list:
        """
        Returns the closest vector IDs to a specific query/text.

        Args:
            text (str): The text to search for.
            n (int): The number of results to return.

        Returns:
            list: The IDs of the vectors (in order of closest to farthest).
        """

        q_response = self.openai_client.embeddings.create(
            input=[text], model="text-embedding-3-small"
        )
        q_embedding = q_response.data[0].embedding
        query_array = np.array([q_embedding])

        D, I = self.db["embeddings"].search(query_array, n)
        return list(I[0])

    def query_min_date(
        self, text: str, min_date: datetime, n: int = 10, date_field: str = "datetime"
    ) -> list:
        """
        Returns the closest vector IDs to a specific query/text, with a minimum date filter.

        Args:
            text (str): The text to search for.
            min_date (datetime): The minimum date to search from.
            n (int): The number of results to return.
            date_field: The field in the metadata to use for the date.

        Returns:
            list: The IDs of the vectors (in order of closest to farthest).
        """

        all_indices_sorted = self.query(text, self.count())

        return_me = []
        for i in range(0, len(all_indices_sorted)):
            if self.db["metadata"][all_indices_sorted[i]][date_field] >= min_date:
                return_me.append(all_indices_sorted[i])

            if len(return_me) >= n:
                break

        return return_me


# TODO If this works, it should be an agent with setllm() supported, etc.
# TODO Right now, we don't actually save sources. It's an important feature (track reliability, etc. too!) but we want to ensure the POC works well first.
class FactRAGFileCache:

    def __init__(
        self,
        folder_path: str,
        openai_api_key: str,
        cache_file: str = "cache.json",
        rag_db_file="vector_db.pickle",
        crawler=None,
        chunker=None,
    ) -> None:
        """
        This is a RAG-based fact database. We build a database of facts available in JSON and via RAG and use this as a basic search engine for information. We use our own DB to index all facts, but also maintain a list of facts, sources, etc. in a JSON file. Finally, we keep a cache of all content and assume URLs do not get updated; we'll change this process in the future.

        Args:
            folder_path (str): The folder where everything will be stored.
            openai_api_key (str): The OpenAI API key. Used for RAG embeddings.
            cache_file (str, optional): The name of the cache file. Defaults to "cache.json".
            rag_db_folder (str, optional): The folder where the database will be stored. Defaults to "cdb".
            crawler (optional): The crawler to use. Defaults to None, in which case a Playwright crawler will be used.
            chunker (optional): The sort of chunker to use. Defaults to None, in which case a GPT-4 chunker will be used.
        """
        self.root_path = folder_path
        self.root_parsed = os.path.join(folder_path, "parsed")
        self.root_original = os.path.join(folder_path, "original")
        self.cache_file = os.path.join(folder_path, cache_file)
        self.rag_db_file = os.path.join(folder_path, rag_db_file)
        self.openai_api_key = openai_api_key

        if chunker is None:
            self.chunker = ChunkerGPT4(openai_api_key)
        else:
            self.chunker = chunker

        # Use the same default crawler for all other agents.
        # TODO Eventually we'll want to have an array agents we use to get content.
        if crawler is None:
            self.crawler = crawlerPlaywright()
        else:
            self.crawler = crawler

        # Set up / load cache
        self.cache = self.load_cache()

        # Vector DB.
        self.vector_db = VectorDBDict(self.rag_db_file, self.openai_api_key)

    def get_facts_as_dict(self, n_results=-1, min_date: datetime = None) -> list:
        """
        Get all facts as a list.

        Args:
            n_results (int, optional): The number of results to return. Defaults to -1, in which case all results are returned.

        Returns:
            list: A list of fact dictionaries containing content, source, and added (the date string for when the fact was added).
        """

        all_facts_raw = self.vector_db.db["original_texts"]
        all_metadata_raw = self.vector_db.db["metadata"]

        facts = {}
        for i in range(0, len(all_facts_raw)):
            fact_id = f"f{i}"
            fact_content = all_facts_raw[i]
            fact_source = all_metadata_raw[i]["source"]
            fact_datetime_string = all_metadata_raw[i]["datetime"].strftime(
                "%Y-%m-%d %H:%M:%S"
            )
            fact_timestamp = all_metadata_raw[i]["added_on_timestamp"]
            fact_datetime = all_metadata_raw[i]["datetime"]

            add_fact = False
            if min_date is None:
                add_fact = True
            else:
                if min_date <= fact_datetime:
                    add_fact = True

            if add_fact:
                facts[fact_id] = {
                    "content": fact_content,
                    "source": fact_source,
                    "added": fact_datetime_string,
                    "added_on_timestamp": fact_timestamp,
                }

        return facts

    def get_facts_as_list(self) -> list:
        """
        Get all facts as a list.

        Args:
            None

        Returns:
            list: A list of facts (as strings).
        """

        return self.vector_db.db["original_texts"].copy()

    def count_facts(self) -> int:
        """
        Returns the number of facts in the knowledge database.

        Returns:
            int: The number of facts in the knowledge database.
        """
        return self.vector_db.count()

    def get_fact_details(self, fact_id: str) -> dict:
        """
        Returns similar structure as query_to_fact_list() but for a specific fact ID. Returns NONE otherwise.

        Args:
            fact_id (str): The fact ID to get.

        Returns:
            dict: A dictionary with the content, source, added date, and added timestamp.
        """

        f_str = fact_id.lower()
        if f_str[0] == "f":
            f_id_as_int = int(f_str[1:])
        else:
            f_id_as_int = int(f_str)

        text, metadata = self.vector_db.get(f_id_as_int)

        fact_content = text
        fact_source = metadata["source"]
        fact_datetime = metadata["datetime"].strftime("%Y-%m-%d %H:%M:%S")
        fact_timestamp = metadata["added_on_timestamp"]

        return {
            "content": fact_content,
            "source": fact_source,
            "added": fact_datetime,
            "added_on_timestamp": fact_timestamp,
        }

    def query_to_fact_list(
        self, query: str, n_results: int = 10, since_date: datetime = None
    ) -> dict:
        """
        Takes a query and finds the closest semantic matches to the query in the knowledge base.

        Args:
            query (str): The query to search for.
            n_results (int, optional): The number of results to return. Defaults to 10.
            since_date (datetime, optional): The date to search from. Defaults to None, in which case all dates are searched.

        Returns:
            dict: A list of the facts found, with the key being the fact ID and each fact having its source, add date, and content info.
        """

        r = []

        if n_results == -1:
            n_results = self.count_facts()

        if since_date is None:

            result_ids = self.vector_db.query(query, n_results)

        else:

            result_ids = self.vector_db.query_min_date(
                query, since_date, n_results, "datetime"
            )

        facts = {}
        for f_id_as_int in result_ids:
            fact_id = f"f{f_id_as_int}"

            text, metadata = self.vector_db.get(f_id_as_int)

            fact_content = text
            fact_source = metadata["source"]
            fact_datetime = metadata["datetime"].strftime("%Y-%m-%d %H:%M:%S")
            fact_timestamp = metadata["added_on_timestamp"]

            facts[fact_id] = {
                "content": fact_content,
                "source": fact_source,
                "added": fact_datetime,
                "added_on_timestamp": fact_timestamp,
            }

        return facts

    def query_to_fact_content(
        self, query: str, n_results: int = 10, since_date=None, skip_separator=False
    ) -> str:
        """
        Takes a query and finds the closest semantic matches to the query in the knowledge base.

        Args:
            query (str): The query to search for.
            n_results (int, optional): The number of results to return. Defaults to 10.
            since_date ([type], optional): The date to search from. Defaults to None, in which case all dates are searched.
            skip_separator (bool, optional): Whether to prepend and append a note horizontal line and title to the string being returned. Defaults to False.

        Returns:
            str: The content of the facts found, along with the fact IDs.

        """

        facts = self.query_to_fact_list(query, n_results, since_date)

        if len(facts) == 0:
            return ""

        fact_content = ""
        if not skip_separator:
            fact_content = """--- START FACTS ---------------------------\n"""

        for key, fact in facts.items():
            fact_content += key + ": " + fact["content"] + "\n"

        if not skip_separator:
            fact_content += """--- END FACTS ---------------------------\n"""

        return fact_content

    def get_all_recent_facts(self, days: float = 1, skip_separator=False) -> str:
        """
        Returns a list of all facts and sources added in the last n days.

        Args:
            days (float, optional): The number of days to search back. Defaults to 1. Can be fractional as well.
            skip_separator (bool, optional): Whether to prepend and append a note horizontal line and title to the string being returned. Defaults to False.

        Returns:
            str: The content of the facts found, along with the fact IDs.
        """

        fact_content = ""

        if not skip_separator:
            fact_content = """--- START FACTS ---------------------------\n"""

        # min_date_timestamp = (datetime.now() - timedelta(days=days)).timestamp()
        # applicable_facts = self.query_to_fact_list("", -1, min_date_timestamp)

        min_date = datetime.now() - timedelta(days=days)
        min_date_timestamp = min_date.timestamp()

        # This "" approach doesn't work because OpenAI errors out.
        # applicable_facts = self.query_to_fact_list("", -1, min_date)
        applicable_facts = self.get_facts_as_dict(-1, min_date)

        for key, fact in applicable_facts.items():
            if fact["added_on_timestamp"] > min_date_timestamp:
                fact_content += key + ": " + fact["content"] + "\n"

        if not skip_separator:
            fact_content += """--- END FACTS ---------------------------\n"""

        return fact_content

    def get_fact_source(self, fact_id: str) -> str:
        """
        Returns the source of a fact given its ID.

        Args:
            fact_id: The fact ID to get.

        Returns:
            str: The source of the fact.
        """

        f_str = fact_id.lower()
        if f_str[0] == "f":
            f_id_as_int = int(f_str[1:])
        else:
            f_id_as_int = int(f_str)

        text, metadata = self.vector_db.get(f_id_as_int)

        return metadata["source"]

    def get_fact_content(self, fact_id: str) -> str:
        """
        Returns the content of a fact given its ID.

        Args:
            fact_id: The fact ID to get.

        Returns:
            str: The content of the fact.
        """

        f_str = fact_id.lower()
        if f_str[0] == "f":
            f_id_as_int = int(f_str[1:])
        else:
            f_id_as_int = int(f_str)

        text, metadata = self.vector_db.get(f_id_as_int)

        return text

    def add_fact(self, fact: str, url: str) -> bool:
        """
        Adds a fact to the knowledge base.

        Args:
            fact (str): The fact to add.
            url (str): The URL source of the fact.

        Returns:
            bool: True if the fact was added, False otherwise.
        """
        return self.add_facts([fact], [url])

    def add_facts(self, facts: list, sources: list) -> bool:
        """
        Adds a facts to the knowledge base.

        Args:
            facts (list): List of strings. Each string is a fact.
            sources (list): List of sources (e.g., URLs) for the facts.

        Returns:
            bool: True if the facts were added, False otherwise.
        """

        # fact_id_start = self.facts_rag_collection.count() + 1

        added_now = datetime.now()
        added_now_timestamp = added_now.timestamp()
        # added_now_string = added_now.strftime("%Y-%m-%d %H:%M:%S")

        # fact_ids = []
        metadatas = []
        for i in range(0, len(facts)):
            # fact_ids.append(f"f{fact_id_start + i}")
            metadatas.append(
                {
                    "added_on_timestamp": added_now_timestamp,
                    "datetime": added_now,
                    "source": sources[i],
                }
            )

        self.vector_db.add_texts(facts, metadatas)
        self.vector_db.save()

        return True

    def facts_from_url(self, url: str, topic: str) -> None:
        """
        Given a URL, extract facts from it and save them to our DB and the facts dictionary. Also returns the facts in an array, in case one wants to analyze new facts.

        Args:
            url (str): Location of the content.
            topic (str): a brief description of the research you are undertaking.
        """

        content = self.get(url)

        facts = self.chunker.chunk(content, topic)
        sources = []
        for f in facts:
            sources.append(url)

        if len(facts) > 0:
            return self.add_facts(facts, sources)
        return False

    # This builds facts based on RSS feeds.
    def new_get_rss_links(self, rss_url, topic) -> None:
        """
        Crawls an RSS feed and its posts.

        Args:
            rss_url (str): The URL of the RSS feed.
            topic (str): a brief description of the research you are undertaking.
        """

        rss_agent = RSSAgent(rss_url, crawler=self.crawler)
        urls = rss_agent.get_news_as_list()

        for url in urls:
            if not self.in_cache(url):
                print("RSS RESULT: " + url)
                # try:
                self.facts_from_url(url, topic)
                # except:
                #    print("Error; failed to get content from " + url)

    # Builds a fact base baed on news from Bing.
    def new_get_new_bing_news(
        self, api_key, subscription_endpoint, topic, queries
    ) -> None:
        """
        Uses Bing to get recent news on the queries and associated topics.

        Args:
            api_key (str): The Bing API key.
            subscription_endpoint (str): The Bing subscription endpoint.
            topic (str): a brief description of the research you are undertaking.
            queries (list[str]): A list of queries to search for.
        """

        news_agent = NewsBingAgent(api_key, subscription_endpoint)
        for q in queries:
            results_urls = news_agent.get_news_as_list(q)
            for url in results_urls:
                if not self.in_cache(url):
                    print("NEWS RESULT: " + url)
                    self.facts_from_url(url, topic)

    # This builds facts based on news articles.
    def new_get_new_info_news(
        self,
        newsapi_api_key,
        topic,
        queries,
        top_headlines=False,
    ) -> None:
        """
        Uses the News API to find new information and extract facts from it.

        Args:
            newsapi_api_key (str): The News API key.
            topic (str): a brief description of the research you are undertaking.
            queries (list[str]): A list of queries to search for.
            top_headlines (bool, optional): Whether to search for top headlines. Defaults to False.
        """

        news_agent = NewsAPIAgent(
            newsapi_api_key, top_headlines=top_headlines, crawler=self.crawler
        )

        for q in queries:
            results = news_agent.get_news_as_list(q)
            for result in results["articles"]:
                url = result["url"]
                if not self.in_cache(url):
                    print("NEWS RESULT: " + url)
                    self.facts_from_url(url, topic)

    # POC for FT
    def get_ft_news(self, ft_user, ft_pass, topic) -> None:
        """
        Uses the Financial Times Agent to find new information and extract facts from it.

        Args:
            ft_user (str): The Financial Times username.
            ft_pass (str): The Financial Times password.
            topic (str): a brief description of the research you are undertaking.
        """

        fta = FinancialTimesAgent(ft_user, ft_pass)
        urls, html_content, text_content = fta.get_news()

        if len(urls) != len(text_content):
            raise ValueError("URLs and text content are not the same length.")

        for i in range(0, len(urls)):
            url = urls[i]
            content = text_content[i]

            if not self.in_cache(url):
                print("FT RESULT: " + url)
                self.force_content(url, content)
                self.facts_from_url(url, topic)

    # This builds facts based on all the google searches.
    def new_get_new_info_google(
        self,
        google_api_key,
        google_search_id,
        google_search_queries,
        topic,
    ) -> None:
        """
        Uses Google search to find new information and extract facts from it.

        Args:
            google_api_key (str): The Google API key.
            google_search_id (str): The Google search ID.
            google_search_queries (list[str]): A list of queries to search for.
            topic (str): a brief description of the research you are undertaking.
        """

        self.google_api_key = google_api_key
        self.google_search_id = google_search_id
        self.google_search_queries = google_search_queries

        webagent = WebSearchAgent(api_key=self.google_api_key)

        for google_search_query in self.google_search_queries:

            try:
                results = webagent.search_google(
                    query=google_search_query,
                    custom_search_engine_id=self.google_search_id,
                    num=_DEFAULT_NUM_SEARCH_RESULTS,
                )
            except:
                print("Error; failed to get content from " + google_search_query)
                continue

            try:
                for result in results:
                    pass
            except:
                print(
                    "Error; failed to get 'result' in 'results' from "
                    + google_search_query
                )
                continue

            for result in results:
                if not self.in_cache(result.url):
                    try:
                        print("SEARCH RESULT: " + result.url)
                        # page_content = self.get(result.url)
                        self.facts_from_url(result.url, topic)
                        # print(page_content)
                    except Exception as e:
                        print(f"Failed to get content from {result.url}\n{e}")

    def save_state(self) -> None:
        """
        Saves the in-memory changes to the knowledge base to the JSON cache file.
        """
        with open(self.cache_file, "w") as f:
            json.dump(self.cache, f, cls=DjangoJSONEncoder)

    def load_cache(self) -> None:
        """
        Loads the cache from the cache file, or creates the relevant files and folders if one does not exist.
        """

        if not os.path.exists(self.root_path):
            os.makedirs(self.root_path)

        if not os.path.exists(self.root_parsed):
            os.makedirs(self.root_parsed)

        if not os.path.exists(self.root_original):
            os.makedirs(self.root_original)

        if not os.path.exists(self.cache_file):
            with open(self.cache_file, "w") as f:
                f.write("{}")

        with open(self.cache_file, "r") as f:
            return json.load(f)

    def in_cache(self, uri: str) -> bool:
        """
        Checks if a URI is in the cache already.

        Args:
            uri (str): The URI to check.

        Returns:
            bool: True if the URI is in the cache, False otherwise.
        """
        if uri in self.cache:
            return True
        return False

    def update_cache(
        self, uri: str, obtained_on: datetime, last_accessed: datetime
    ) -> None:
        """
        Updates the cache file for a given URI, specifically when it was obtained and last accessed.

        Args:
            uri (str): The URI to update.
            obtained_on (datetime): The date and time when the content was obtained.
            last_accessed (datetime): The date and time when the content was last accessed.
        """
        uri_md5 = uri_to_local(uri)
        self.cache[uri] = {
            "obtained_on": obtained_on,
            "last_accessed": last_accessed,
            "accessed": 0,
            "uri_md5": uri_md5,
        }
        self.save_state()

    def log_access(self, uri: str) -> None:
        """
        Saves the last accessed time and updates the accessed tracker for a given URI.

        Args:
            uri (str): The URI to update.
        """
        self.cache[uri]["last_accessed"] = datetime.now()
        self.cache[uri]["accessed"] = 1
        self.save_state()

    def get_unaccessed_content(self) -> list[str]:
        """
        Returns a list of URIs that have not been accessed by the agent.

        Returns:
            list[str]: A list of URIs that have not been accessed by the agent.
        """
        unaccessed = []
        for uri in self.cache:
            if self.cache[uri]["accessed"] == 0:
                unaccessed.append(uri)
        return unaccessed

    def force_content(self, uri: str, content: str, check_exists: bool = True) -> bool:
        """
        Forces a specific URI to have specific content (both HTML and text content). Used to fill old links that we don't actually want to crawl.

        Args:
            uri (str): The URI to force content for.
            content (str): The content to force.
            check_exists (bool): checks if content has already been included in the cache before forcing the new content.

        Returns:
            bool: True if the content was forced, False otherwise.
        """

        # If the content already exists and we avoid overwrites, then we don't want to overwrite it.
        if check_exists and self.in_cache(uri):
            return False

        uri_md5 = uri_to_local(uri)
        with open(os.path.join(self.root_original, uri_md5), "w") as f:
            f.write(content)
        with open(os.path.join(self.root_parsed, uri_md5), "w") as f:
            f.write(content)

        self.update_cache(uri, datetime.now(), datetime.now())
        self.log_access(uri)

        return True

    def get(self, uri: str) -> str:
        """
        Returns the content for a given URI. If the content is not in the cache, it will be scraped and added to the cache.

        Args:
            uri (str): The URI to get the content for.

        Returns:
            str: The content for the given URI.
        """
        uri_md5 = uri_to_local(uri)
        if uri in self.cache:
            with open(os.path.join(self.root_parsed, uri_md5), "r") as f:
                return f.read()
        else:
            # scraper = WebpageAgent()

            # content_raw = scraper.scrape(uri, text_only=False, body_only=False)
            # with open(os.path.join(self.root_original, uri_md5), "w") as f:
            #    f.write(content_raw)

            # content_parsed = scraper.scrape(uri, text_only=True, body_only=True)
            # with open(os.path.join(self.root_parsed, uri_md5), "w") as f:
            #    f.write(content_parsed)

            try:
                content, text = self.crawler.get_content(uri)
            except Exception as e:
                print(f"Failed to get content from {uri}\n{e}")
                content = ""
                text = ""

            with open(os.path.join(self.root_original, uri_md5), "w") as f:
                f.write(content)
            with open(os.path.join(self.root_parsed, uri_md5), "w") as f:
                f.write(text)

            self.update_cache(uri, datetime.now(), datetime.now())

            return text

    def add_content(self, content: str, uri: str = None) -> None:
        """
        Adds content to cache.

        Args:
            content (str): The content to add to the cache.
            uri (str, optional): The URI to use for the content. Defaults to None, in which case an MD5 sum of the content will be used.
        """
        if uri is None:
            uri = hashlib.md5(content.encode("utf-8")).hexdigest()
        uri_md5 = uri_to_local(uri)
        with open(os.path.join(self.root_parsed, uri_md5), "w") as f:
            f.write(content)
        self.update_cache(uri, datetime.now(), datetime.now())

    def add_content_from_file(self, filepath: str, uri: str = None) -> None:
        """
        Adds content from a text file to the cache.

        Args:
            filepath (str): The path to the file to add to the cache.
            uri (str, optional): The URI to use for the content. Defaults to None, in which case an MD5 sum of the content will be used.
        """
        with open(filepath, "r") as f:
            content = f.read()
        self.add_content(content, uri)


class FactBot:

    def __init__(
        self,
        knowledge_db: FactRAGFileCache,
        openai_api_key: str = None,
        chatbot: ChatBot = None,
    ) -> None:
        """
        The FactBot is like a ChatBot but enables you to ask questions that reference an underlying RAG database (KnowledgeBaseFileCache), which then enables the chatbot to cite sourcable facts.

        Args:
            knowledge_db (FactRAGFileCache): The knowledge database to use.
            openai_api_key (str, optional): The OpenAI API key. Defaults to None.
            chatbot (ChatBot, optional): The PhaseLLB chatbot to use. Defaults to None, in which case an OpenAI chatbot is used (and the OpenAI API key must be provided).
        """
        if openai_api_key is None and chatbot is None:
            raise ValueError("One of openai_api_key or chatbot must be provided.")

        if chatbot is not None:
            self.chatbot = chatbot
        else:
            llm = OpenAIGPTWrapper(openai_api_key, model="gpt-4-turbo-preview")
            self.chatbot = ChatBot(llm)
            self.chatbot.messages = [
                {"role": "system", "content": system_prompt_question_continuous}
            ]

        self.knowledge_db = knowledge_db

    def ask(self, question: str, clean_sources: bool = True) -> str:
        """
        Ask a question to the FactBot. This will query the underlying knowledge database and use the returned facts to answer the question.

        Args:
            question (str): The question to ask.
            clean_sources (bool, optional): Whether to clean the sources in the response. Defaults to True; in this case, it will replace fact IDs with relevant source links at the end of the response.

        Returns:
            str: The response to the question.
        """
        message = self.knowledge_db.query_to_fact_content(question) + "\n\n" + question
        response = self.chatbot.chat(message)
        if clean_sources:
            return clean_fact_citations(self.knowledge_db, response)
        else:
            return response

    def source(self, fact_id: str) -> str:
        """
        Returns the URL source for a given fact ID.

        Args:
            fact_id (str): The fact ID to get the source for.

        Returns:
            str: The URL source for the given fact ID.
        """
        return self.knowledge_db.get_fact_source(fact_id)

    def clean_and_source_to_html(
        self, text_to_clean: str, start_count: int = 0
    ) -> list:
        """
        Returns a formatted response with sourced HTML. This is used for emergingtrajectories.com and acts as a base for anyone else wanting to build similar features.

        Args:
            text_to_clean: The text to clean/cite/source.
            start_count: The starting count for the sources.

        Returns:
            list: two strings -- the actual response in the first case, and the sources in the second case, and an integer representing the new source count.
        """

        pattern = r"\[f[\d\s\,f]+\]"
        new_text = ""
        sources_text = ""
        ref_ctr = start_count
        last_index = 0

        for match in re.finditer(pattern, text_to_clean, flags=re.IGNORECASE):

            if match.group(0).find(",") == -1:
                ref_ctr += 1
                ref = match.group(0)[1:-1].strip()
                ref = ref.lower()

                new_text += text_to_clean[last_index : match.start()]
                new_text += f"""<a class='source_link' target='_blank' href='{self.source(ref)}'>{ref_ctr}</a>"""

                # Save the source
                fact_text = self.knowledge_db.get_fact_content(ref)
                new_source_text = f"""<span class='fact_span'><b>{ref_ctr}:</b> {fact_text} <a href='{self.source(ref)}' target='_blank'>View Source</a></span>"""
                sources_text += new_source_text + "\n"

                last_index = match.end()
            else:
                refs = match.group(0)[1:-1].split(",")
                ref_arr = []
                ref_str = ""
                for ref in refs:
                    ref = ref.strip()
                    ref = ref.lower()
                    ref_ctr += 1
                    ref_arr.append(str(ref_ctr))

                    # Add the source to the text
                    new_text_source_num = f"""<a class='source_link' target='_blank' href='{self.source(ref)}'>{ref_ctr}</a>"""
                    ref_str += " " + new_text_source_num

                    # Save the source
                    fact_text = self.knowledge_db.get_fact_content(ref)
                    new_source_text = f"""<span class='fact_span'><b>{ref_ctr}:</b> ${fact_text} <a href='{self.source(ref)}' target='_blank'>View Source</a></span>"""
                    sources_text += new_source_text + "\n"

                new_text += text_to_clean[last_index : match.start()] + ref_str
                last_index = match.end()

        new_text += text_to_clean[last_index:]

        return new_text, sources_text, ref_ctr


def clean_fact_citations(knowledge_db: FactRAGFileCache, text_to_clean: str) -> str:
    """
    Converts fact IDs referenced in a piece of text to relevant source links, appending sources as end notes in the document/text.

    Args:
        knowledge_db (FactRAGFileCache): The knowledge database to use for fact lookups.
        text_to_clean (str): The text to clean.

    Returns:
        str: The cleaned text.
    """
    bot = FactBot(knowledge_db, knowledge_db.openai_api_key)
    pattern = r"\[f[\d\s\,f]+\]"
    new_text = ""
    ref_ctr = 0
    last_index = 0
    sources_list = ""
    for match in re.finditer(pattern, text_to_clean):
        if match.group(0).find(",") == -1:
            ref_ctr += 1
            ref = match.group(0)[1:-1].strip()
            new_text += text_to_clean[last_index : match.start()]
            new_text += f"[{ref_ctr}]"
            sources_list += f"{ref_ctr} :: " + bot.source(f"{ref}") + "\n"
            last_index = match.end()
        else:
            refs = match.group(0)[1:-1].split(",")
            ref_arr = []
            for ref in refs:
                ref = ref.strip()
                ref_ctr += 1
                ref_arr.append(str(ref_ctr))
                sources_list += f"{ref_ctr} :: " + bot.source(f"{ref}") + "\n"
            ref_str = "[" + ", ".join(ref_arr) + "]"
            new_text += text_to_clean[last_index : match.start()] + ref_str
            last_index = match.end()

    new_text += text_to_clean[last_index:]

    if ref_ctr == 0:
        return text_to_clean
    else:
        return new_text + "\n\nSources:\n" + sources_list



"""
Crawlers provide a standardized approach to interacting with with web pages and extracting information. We have a number of crawlers based on PhaseLLM (Python requests) and ones using Playwright (headlessly and with a front-end) to enable flexible scraping.

All scraping agents return the raw HTML content and the extracted text content.
"""

from playwright.sync_api import sync_playwright
from bs4 import BeautifulSoup

from phasellm.agents import WebpageAgent

from scrapingbee import ScrapingBeeClient


def _bs4_childtraversal(html: str) -> str:
    """
    Recursively travserse the DOM to extract content.

    Args:
        html (str): HTML content

    Returns:
        str: Extracted content
    """

    if len(str(html).strip()) < 2:
        return ""

    new_html = ""

    for content in html:
        contentname = ""

        if isinstance(content, str):
            contentname = ""
        elif content.name is not None:
            contentname = content.name.lower()

        if contentname in ["p", "pre", "h1", "h2", "h3", "h4", "h5", "h6", "span"]:
            text = content.get_text()
            num_words = len(text.strip().split(" "))
            # print(num_words)
            if num_words > 7:
                new_html = new_html + content.get_text() + "\n\n"
        else:
            new_html = new_html + _bs4_childtraversal(content)

    return new_html


def _get_text_bs4(html: str) -> str:
    """
    Extract text content from HTML using BeautifulSoup.

    Args:
        html (str): HTML content

    Returns:
        str: Extracted text content
    """

    new_html = "<html><body>"

    souppre = BeautifulSoup(html, "html.parser")
    soup = souppre.body

    for content in soup.contents:
        contentname = ""
        if content.name is not None:
            contentname = content.name.lower()
        if contentname not in ["script", "style"]:
            new_html = new_html + _bs4_childtraversal(content)

    new_html = new_html + "</body></html>"

    newsoup = BeautifulSoup(new_html, "html.parser")
    text = newsoup.get_text()

    return text


class crawlerPlaywright:

    def __init__(self, headless: bool = True) -> None:
        """
        Crawler that uses Playwright to scrape web pages.

        Args:
            headless (bool, optional): Run the browser in headless mode. Defaults to True.
        """
        self.headless = headless

    def get_content(self, url: str) -> tuple[str, str]:
        """
        Gets content for a specific URL.

        Args:
            url (str): URL to scrape

        Returns:
            tuple[str, str]: Raw HTML content and extracted text content (in this order)
        """

        content = ""
        text = ""
        with sync_playwright() as playwright:

            browser = playwright.chromium.launch(headless=self.headless)
            page = browser.new_page()

            # Navigate to the webpage
            page.goto(url)

            # Extract data
            content = page.content()

            # Close the browser
            browser.close()

        text = _get_text_bs4(content)

        return content, text


class crawlerPhaseLLM:

    def __init__(self):
        """
        PhaseLLM scraper. Uses Python requests and does not execute JS.
        """
        self.scraper = WebpageAgent()

    def get_content(self, url):
        """
        Gets content for a specific URL.

        Args:
            url (str): URL to scrape

        Returns:
            tuple[str, str]: Raw HTML content and extracted text content (in this order)
        """
        content_raw = self.scraper.scrape(url, text_only=False, body_only=False)
        content_parsed = self.scraper.scrape(url, text_only=True, body_only=True)
        return content_raw, content_parsed


class crawlerScrapingBee:

    def __init__(self, api_key: str):
        """
        Crawler that uses ScrapingBee to scrape web pages.
        """
        self.client = ScrapingBeeClient(api_key=api_key)

    def get_content(self, url):
        """
        Gets content for a specific URL.

        Args:
            url (str): URL to scrape

        Returns:
            tuple[str, str]: Raw HTML content and extracted text content (in this order)
        """

        response = self.client.get(url)
        content_raw = response.content.decode("utf-8")
        content_parsed = _get_text_bs4(content_raw)
        return content_raw, content_parsed



"""
chunkers.py is used to chunk facts using different strategies. Emerging Trajectories started by chunking via GPT-4, but we can also appreciate using sentences, paragraphs, or other verbatim approaches. We'll be adding more chunkers as time goes on.

Chunkers should simply take a piece of content and chunk it into a list of facts.
"""

from phasellm.llms import OpenAIGPTWrapper, ChatBot, ChatPrompt
from phasellm.agents import WebpageAgent, WebSearchAgent

fact_system_prompt = """You are a researcher helping extract facts about {topic}, trends, and related observations. We will give you a piece of content scraped on the web. Please extract facts from this. Each fact should stand on its own, and can be several sentences long if need be. You can have as many facts as needed. For each fact, please start it as a new line with "---" as the bullet point. For example:

--- Fact 1... This is the fact.
--- Here is a second fact.
--- And a third fact.

Please do not include new lines between bullet points. Make sure you write your facts in ENGLISH. Translate any foreign language content/facts/observations into ENGLISH.

We will simply provide you with content and you will just provide facts."""


class ChunkerGPT4:

    def __init__(self, openai_api_key: str, model="gpt-4-turbo"):
        """
        Chunker based on GPT-4 reading text and providing a list of facts.

        Args:
            openai_api_key (str): The OpenAI API key.
            model (str): The OpenAI model to use. Defaults to "gpt-4-turbo".
        """
        self.openai_api_key = openai_api_key
        self.model = model

    def chunk(self, content: str, topic: str) -> list[str]:
        """
        Chunk text into facts.

        Args:
            content (str): The content to chunk.
            topic (str): The topic to focus on when building facts.

        Returns:
            list[str]: The list of facts.
        """

        llm = OpenAIGPTWrapper(self.openai_api_key, model=self.model)
        chatbot = ChatBot(llm)
        chatbot.messages = [{"role": "system", "content": fact_system_prompt}]

        prompt_template = ChatPrompt(
            [
                {"role": "system", "content": fact_system_prompt},
            ]
        )

        chatbot.messages = prompt_template.fill(topic=topic)

        response = chatbot.chat(content)

        lines = response.split("\n")

        facts = []

        for line in lines:
            if line[0:4] == "--- ":
                fact = line[4:]
                facts.append(fact)

        return facts


class ChunkerNewLines:

    def __init__(self, min_length: int = 7):
        """
        Chunker using line breaks for content.

        Args:
            min_length (int): The minimum length (in characters) of a fact. Defaults to 7 characters.
        """
        self.min_length = min_length

    def chunk(self, content: str, topic: str = None) -> list[str]:
        """
        Chunk text into facts.

        Args:
            content (str): The content to chunk.
            topic (str): The topic to focus on when building facts. This defaults to None so we can keep the same function calls as other chunkers.

        Returns:
            list[str]: The list of facts.
        """

        lines = content.split("\n")

        facts = []

        for line in lines:
            ls = line.strip()
            if len(ls) >= self.min_length:
                facts.append(ls)

        return facts



import requests
import json
from datetime import datetime
import dateparser
from typing import Union, List
import warnings
import time


def hello() -> None:
    """
    Just a hello() message/function to confirm you've installed everything!
    """
    print("Welcome to the Emerging Trajectories package! We've been expecting you. 😉")


# TODO: document
class Statement(object):
    def __init__(self, title, fill_in_the_blank):
        self.id = -1
        self.title = title
        self.fill_in_the_blank = fill_in_the_blank
        self.description = ""
        self.deadline = None
        self.created_at = None
        self.updated_at = None
        self.created_by = None


# TODO: document
class Forecast(object):

    def __init__(self, title, value, justification):
        self.id = -1
        self.title = title
        self.value = value
        self.justification = justification

        self.statement = None
        self.created_at = None
        self.updated_at = None
        self.created_by = None
        self.prediction_agent = None
        self.additional_data = {}
        self.prior_forecast = None
        self.next_forecasts = []
        self.is_human = False


class Client(object):

    # The base URL for the API, in case we need to change it or if someone wants to self-host anything.
    base_url = "https://app.emergingtrajectories.com/a/api/v0.2/"

    def __init__(self, api_key: str) -> None:
        """
        Launch the Emerging Trajectories Client.

        Args:
            api_key: the API key for the Emerging Trajectories platform.
        """
        self.api_key = api_key

    def create_statement(
        self, title: str, description: str, deadline: datetime, fill_in_the_blank: str
    ) -> dict:
        """
        Create a new statement that users will be forecasting against/for.

        Args:
            title: the title of the statement
            description: a more detailed description of the statement
            deadline: the deadline (date and time), typically when we'll learn the "right answer" for the forecasting process
            fill_in_the_blank: the "fill in the blank" part of the statement, which is what users will be forecasting against

        Returns:
            dict: the newly created statement returned from the platform
        """
        url = self.base_url + "create_statement"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {
            "title": title,
            "description": description,
            "deadline": deadline,
            "fill_in_the_blank": fill_in_the_blank,
        }
        response = requests.post(url, headers=headers, data=json.dumps(data))
        if response.status_code == 201:
            return response.json()
        else:
            raise Exception(response.text)

    def get_statement(self, statement_id: int) -> dict:
        """
        Returns a given statement from the platform. Includes title, description, deadline, and fill-in-the-blank.

        Args:
            statement_id: the ID of the statement to retrieve

        Returns:
            dict: the statement from the platform
        """
        url = self.base_url + "get_statement" + "/" + str(statement_id)
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        response = requests.post(url, headers=headers)
        if response.status_code == 200:
            return response.json()
        else:
            raise Exception(response.text)

    def get_most_recent_forecast(
        self, statement_id: int, prediction_agent: str = None
    ) -> int:
        """
        Returns the most recent forecast for a given statement. This is useful for creating a new forecast that is an extension of a prior forecast.

        Args:
            statement_id: the ID of the statement to retrieve the most recent forecast for
            prediction_agent: the string for a prediction agent, if you want to further filter the most recent forecast

        Returns:
            int: the ID of the most recent forecast for the given statement
        """
        url = self.base_url + "get_most_recent_forecast/" + str(statement_id)
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {}
        if prediction_agent is not None:
            data["prediction_agent"] = prediction_agent
        response = requests.post(url, headers=headers, json=data)
        if response.status_code == 200:
            return int(response.json()["forecast_id"])
        else:
            raise Exception(response.text)

    def get_forecast(self, forecast_id: int) -> dict:
        """
        Returns a specific forecast's details from the platform. This typically includes the forecast title, the value associated with the fill-in-the-blank component of a statement, and justificaiton for the forecast.

        Args:
            forecast_id: the ID of the forecast to retrieve

        Returns:
            dict: the forecast details in the form of a dictionary object
        """
        url = self.base_url + "get_forecast" + "/" + str(forecast_id)
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        response = requests.post(url, headers=headers)
        if response.status_code == 200:
            return response.json()
        else:
            raise Exception(response.text)

    def create_forecast(
        self,
        statement_id: int,
        title: str,
        justification: str,
        value: float,
        prediction_agent: str,
        additional_data: dict = {},
        prior_forecast: int = None,
        is_human: bool = False,
    ) -> None:
        """
        Creates a forecast tied to a specific statement.

        Args:
            statement_id: the ID of the statement to tie the forecast to
            title: the title of the forecast
            justification: the justification for the forecast
            value: the value associated with the fill-in-the-blank component of the statement
            prediction_agent: the agent making the prediction
            additional_data: any additional data to include with the forecast. This is not used anywhere, but can be helpful in audting or researching forecast effectiveness
            prior_forecast: if this forecast is an extension of an earlier forecast, the ID of the prior forecast
            is_human: whether the prediction is human-generated

        Returns:
            dict: the newly created forecast from the platform
        """
        url = self.base_url + "create_forecast/" + str(statement_id)
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {
            "title": title,
            "justification": justification,
            "value": value,
            "prediction_agent": prediction_agent,
            "additional_data": additional_data,
            "is_human": is_human,
        }
        if prior_forecast is not None:
            data["prior_forecast"] = prior_forecast
        response = requests.post(url, headers=headers, data=json.dumps(data))
        if response.status_code == 201:
            return response.json()
        else:
            raise Exception(response.text)


class EmergingTrajectoriesClient(object):
    """
    This is the new and improved Emerging Trajectories client and should replace older clients. We keep the older ones for backwards compatability, for now.

    This client enables the same features as earlier clients, but also provides support for creating fact bases, automations, documents, and more.
    """

    # The base URL for the API, in case we need to change it or if someone wants to self-host anything.
    base_url = "https://emergingtrajectories.com/a/api/v0.2/"

    def __init__(self, api_key: str) -> None:
        """
        Args:
            api_key: the API key for the Emerging Trajectories platform.
        """
        self.api_key = api_key

    def create_factbase(self, title: str, description: str) -> str:
        """
        Create a new factbase on the Emerging Trajectories platform.

        Args:
            title: the title of the factbase
            description: a description of the factbase

        Returns:
            str: The short_code for the factbase.
        """
        url = self.base_url + "create_factbase"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {
            "title": title,
            "description": description,
        }
        response = requests.post(url, headers=headers, data=json.dumps(data))
        if response.status_code == 200:
            r = response.json()
            if "short_code" in r:
                return r["short_code"]
        else:
            raise Exception(response.text)

    def query_factbase(self, short_code, query, days_filter=1, pubdate_days_filter=-1, llm_model=None, llm_temperature=None) -> str:
        """
        Query the fact base as if you'd query a document, but with no document required. The responses here are ephemeral and are not stored anywhere.

        Args:
            short_code: the short code for the factbase to query
            query: the query to run
            days_filter: limit facts to the recent number of days
            pubdate_days_filter: limit facts to the recent number of days based on publication date
            llm_model: the LLM model to use
            llm_temperature: the LLM temperature to use
        Returns:
            str: The response written based on facts in the fact base.
        """

        url = self.base_url + "factbase_query/" + short_code 
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {"query": query}

        data['days_filter'] = days_filter
        data['pubdate_days_filter'] = pubdate_days_filter

        if llm_model is not None:
            data['llm_model'] = llm_model

        if llm_temperature is not None:
            data['llm_temperature'] = llm_temperature

        response = requests.post(url, headers=headers, data=json.dumps(data))

        if response.status_code == 200:
            r = response.json()
            return r['response']
        else:
            raise Exception(response.text)

    def build_research_plan(self, query: str) -> dict:
        """
        Given a research question/query, we create a set of tasks that we can then automate document creation around.

        Args:
            query: the research question/query to build tasks around
        Returns:
            dict: a dictionary with two keys -- "text" which is the text plan, and "plan" which is a JSON plan
        """

        url = self.base_url + "api_research_analyst_build_plan"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {"query": query}

        response = requests.post(url, headers=headers, data=json.dumps(data))

        if response.status_code == 200:
            r = response.json()
            return r
        else:
            raise Exception(response.text)

    def run_research_sub_task(
        self,
        command: str,
        args: Union[str, List[str]],
        document_id: int,
        short_code: str,
        wait_until_completion: bool = True,
    ) -> int:
        """
        Run the research sub-task outlined in a research plan.

        Args:
            command: the command to run
            args: the arguments to pass to the command
            document_id: the ID of the document to attach the research task to
            short_code: the short code for the factbase to use
            wait_until_completion: whether to wait until the task is completed before returning
        Returns:
            int: the ID of the job or -1 if not applicable
        """

        job_id = -1
        job_status = ""

        if command == "NEWS":
            job_id = self.run_data_collector_news(short_code, args)
        elif command == "SEARCH" or command == "WEBSEARCH":
            job_id = self.run_data_collector_serp(short_code, args, 5)
        elif (
            command == "QUERY" or command == "LIST"
        ):  # We should treat lists differently later.
            job_id = self.research_task_header_and_block(document_id, args)

        if wait_until_completion:
            job_status = ""
            while job_status != "complete":
                job_status = self.get_data_collector_job_status(job_id)
                print(job_status)
                time.sleep(10)

        return -1

    # TODO: refactor all run_data_collector_*
    def research_task_header_and_block(self, doc_id: int, query: str) -> int:
        """
        Runs a research analyst task where we generate a header block and then run the query for the document.

        Args:
            doc_id: the ID of the document to attach the research task to
            query: the query to run
        Returns:
            int: the ID of the job
        """

        url = self.base_url + "api_data_collector_run_header_and_block/" + str(doc_id)
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {"query": query}

        response = requests.post(url, headers=headers, data=json.dumps(data))

        if response.status_code == 200:
            r = response.json()
            if "job_id" in r:
                return int(r["job_id"])
        else:
            raise Exception(response.text)

        raise Exception("Failed to create job.")

    def run_data_collector_job(self, factbase_shortcode: str, settings: dict) -> int:
        """
        Creates a new data collector job and runs it. Contact us for information on how to pass settings.

        Args:
            factbase_shortcode: the short code for the factbase
            settings: a dictionary of settings for the data collector job
        Returns:
            int: The ID of the job
        """

        url = self.base_url + "api_data_collector_run_once/" + factbase_shortcode
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {"data_collector_settings": settings}

        response = requests.post(url, headers=headers, data=json.dumps(data))

        if response.status_code == 200:
            r = response.json()
            if "job_id" in r:
                return int(r["job_id"])
        else:
            raise Exception(response.text)

        raise Exception("Failed to create job.")

    # TODO: refactor all run_data_collector_*
    def run_data_collector_serp(
        self,
        factbase_shortcode: str,
        query: Union[str, List[str]],
        n: int = 5,
        settings: dict = None,
    ) -> int:
        """
        Creates a new data collector job and runs it. Contact us for information on how to pass settings.

        Args:
            factbase_shortcode: the short code for the factbase
            query: a string to search or an array of strings
            settings: a dictionary of settings for the data collector job
        Returns:
            int: The ID of the job
        """

        if settings is None:
            settings = {}
        settings_clean = settings.copy()
        settings_clean["collector"] = "SEARCH"
        if query is type(str):
            settings_clean["query"] = [query]
        else:
            settings_clean["query"] = query

        if n > 10:
            settings_clean["n"] = 10
            warnings.warn("The maximum number of results is 10.")
        else:
            settings_clean["n"] = n

        url = self.base_url + "api_data_collector_run_serp/" + factbase_shortcode
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {"data_collector_settings": settings_clean}

        response = requests.post(url, headers=headers, data=json.dumps(data))

        if response.status_code == 200:
            r = response.json()
            if "job_id" in r:
                return int(r["job_id"])
        else:
            raise Exception(response.text)

        raise Exception("Failed to create job.")

    # TODO: refactor all run_data_collector_*
    def run_data_collector_news(
        self,
        factbase_shortcode: str,
        query: Union[str, List[str]],
        n: int = 5,
        settings: dict = None,
    ) -> int:
        """
        Creates a new data collector job and runs it. Contact us for information on how to pass settings.

        Args:
            factbase_shortcode: the short code for the factbase
            query: a string to search or an array of strings
            settings: a dictionary of settings for the data collector job
        Returns:
            int: The ID of the job
        """

        if settings is None:
            settings = {}
        settings_clean = settings.copy()
        settings_clean["collector"] = "NEWSSEARCH4"
        if query is type(str):
            settings_clean["query"] = [query]
        else:
            settings_clean["query"] = query

        if n > 10:
            settings_clean["n"] = 10
            warnings.warn("The maximum number of results is 10.")
        else:
            settings_clean["n"] = n

        url = self.base_url + "api_data_collector_run_news/" + factbase_shortcode
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {"data_collector_settings": settings_clean}

        response = requests.post(url, headers=headers, data=json.dumps(data))

        if response.status_code == 200:
            r = response.json()
            if "job_id" in r:
                return int(r["job_id"])
        else:
            raise Exception(response.text)

        raise Exception("Failed to create job.")

    def get_data_collector_job_status(self, job_id: int) -> str:
        """
        Get the status of a data collector job.

        Args:
            job_id: the ID of the job
        Returns:
            str: The status of the job
        """

        if job_id <= 0 or job_id is None or type(job_id) is not int:
            return (
                "complete"  # We don't have a job ID, so we'll just say it's complete.
            )

        url = self.base_url + "api_data_collector_run_once_status/" + str(job_id)
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }

        response = requests.post(url, headers=headers)

        if response.status_code == 200:
            r = response.json()
            if "job_status" in r:
                return r["job_status"]
        else:
            raise Exception(response.text)

        raise Exception("Failed; unknown error.")

    def create_document(
        self,
        factbase_shortcode: str,
        title: str = None,
        days_filter=99999,
        temperature=None,
        llm=None,
        facts_min_date=None,
        facts_max_date=None,
    ) -> int:
        """
        Create a new document on the Emerging Trajectories platform.

        Args:
            factbase_shortcode: the short code for the factbase to attach the document to
            title (optional): the title of the document
            days_filter (optional): limit facts to the recent number of days

        Returns:
            int: The ID of the document.
        """

        url = self.base_url + "api_doc_create"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {
            "title": title,
            "factbase_shortcode": factbase_shortcode,
            "days_filter": days_filter,
        }

        if temperature is not None:
            data["temperature"] = temperature

        if llm is not None:
            data["llm"] = llm

        if facts_max_date is not None:
            data["facts_max_date"] = facts_max_date.isoformat()

        if facts_min_date is not None:
            data["facts_min_date"] = facts_min_date.isoformat()

        response = requests.post(url, headers=headers, data=json.dumps(data))

        if response.status_code == 200:
            r = response.json()
            if "doc_id" in r:
                return int(r["doc_id"])
        else:
            raise Exception(response.text)

        raise Exception("Failed to create document.")

    def convert_facts_in_text(self, factbase_shortcode: str, text: str):
        """
        Convert text with facts to HTML code with associated links.

        Args:
            factbase_shortcode: the short code for the factbase to attach the document to
            text: the text to convert

        Returns:
            str: the converted text as HTML
            array: a list facts found in the text, with the key being the source ID in the HTML
        """

        url = self.base_url + "api_convert_facts_in_text"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {"factbase_shortcode": factbase_shortcode, "text": text}

        response = requests.post(url, headers=headers, data=json.dumps(data))

        if response.status_code == 200:
            r = response.json()
            return r["html"], r["sources"]
        else:
            raise Exception(response.text)

    def update_document(
        self,
        doc_id: int,
        title: str = None,
        is_public: bool = None,
        short_code: str = None,
    ) -> bool:
        """
        Update document settings.

        Args:
            doc_id: the ID of the document
            title: the new title of the document
            is_public: whether the document is public (updated setting)
            short_code: the short code for the document (updated setting)

        Returns:
            bool: True if successful, False otherwise
        """

        url = self.base_url + "api_doc_update"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {"doc_id": doc_id}

        if title is not None:
            data["title"] = title
        if is_public is not None:
            data["is_public"] = is_public
        if short_code is not None:
            data["short_code"] = short_code

        response = requests.post(url, headers=headers, data=json.dumps(data))

        if response.status_code == 200:
            return True
        else:
            raise Exception(response.text)

    def get_document(self, doc_id: int) -> dict:
        """
        Get document data.

        Args:
            doc_id: the ID of the document

        Returns:
            int: The ID of the document.
        """

        url = self.base_url + "api_doc_get"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {"doc_id": doc_id}

        response = requests.post(url, headers=headers, data=json.dumps(data))

        if response.status_code == 200:
            r = response.json()
            return r
        else:
            raise Exception(response.text)

    def add_viewer(self, doc_id: int, viewer_email: str) -> bool:
        """
        Add a viewer to a private document.

        Args:
            doc_id: the ID of the document
            viewer_email: the email of the viewer to add
        Returns:
            bool: True if successful, False otherwise
        """

        url = self.base_url + "api_doc_add_viewer"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {"doc_id": doc_id, "viewer": viewer_email}

        response = requests.post(url, headers=headers, data=json.dumps(data))

        if response.status_code == 200:
            return True
        else:
            raise Exception(response.text)

    def remove_viewer(self, doc_id: int, viewer_email: str) -> bool:
        """
        Remove a viewer from a private document.

        Args:
            doc_id: the ID of the document
            viewer_email: the email of the viewer to remove
        Returns:
            bool: True if successful or if the email is NOT a viewer already, False otherwise
        """

        url = self.base_url + "api_doc_remove_viewer"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {"doc_id": doc_id, "viewer": viewer_email}

        response = requests.post(url, headers=headers, data=json.dumps(data))

        if response.status_code == 200:
            return True
        else:
            raise Exception(response.text)

    def get_block(self, doc_id: int, block_named_id: str) -> str:
        """
        Get the content of a named block from a document.

        Args:
            doc_id: the ID of the document
            block_named_id: the named ID of the block to retrieve
        Returns:
            str: the content of the block
        """

        url = self.base_url + "api_block_get"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {"doc_id": doc_id, "block_named_id": block_named_id}

        response = requests.post(url, headers=headers, data=json.dumps(data))

        if response.status_code == 200:
            r = response.json()["content"]
            return r
        else:
            raise Exception(response.text)

    def append_header_block(self, doc_id, text: str, is_hidden: bool = False) -> bool:
        """
        Add a header block to a document.

        Args:
            doc_id: the ID of the document to append the header block to
            text: the text (header) to append
            is_hidden: whether the block is hidden in public documents
        Returns:
            bool: True if successful, False otherwise
        """

        url = self.base_url + "api_doc_append_header_block"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {
            "doc_id": doc_id,
            "text": text,
        }

        if is_hidden:
            data["is_hidden"] = "true"

        response = requests.post(url, headers=headers, data=json.dumps(data))

        if response.status_code == 200:
            return True

        return False

    def append_text_block(self, doc_id: int, text: str) -> bool:
        """
        Add a text block to a document.

        Args:
            doc_id: the ID of the document to append the text block to
            text: the text to append
        Returns:
            bool: True if successful, False otherwise
        """

        url = self.base_url + "api_doc_append_text_block"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {
            "doc_id": doc_id,
            "text": text,
        }

        response = requests.post(url, headers=headers, data=json.dumps(data))

        if response.status_code == 200:
            return True

        return False

    def append_hidden_block(self, doc_id: int, prompt: str) -> bool:
        """
        Add a text block to a document.

        Args:
            doc_id: the ID of the document to append the text block to
            prompt: the text (typically a prompt) to append
        Returns:
            bool: True if successful, False otherwise
        """

        url = self.base_url + "api_doc_append_hidden_block"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {
            "doc_id": doc_id,
            "text": prompt,
        }

        response = requests.post(url, headers=headers, data=json.dumps(data))

        if response.status_code == 200:
            return True

        return False

    def append_askai_fts_block(
        self,
        doc_id: int,
        query: str,
        fts_term: str,
        named_id: str = None,
        is_hidden: bool = False,
    ) -> dict:
        """
        Add an AI (FTS) block to a document.

        Args:
            doc_id: the ID of the document to append the text block to
            query: the query to use for the AI block
            fts_term: the full text search (FTS) terms/query to use for the AI block
            named_id: the named ID of the AI block
            is_hidden: whether the block is hidden in public documents
        Returns:
            JSON dict with document information
        """
        url = self.base_url + "api_doc_append_askai_fts_block"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {
            "doc_id": doc_id,
            "query": query,
            "arg": fts_term,
        }

        if is_hidden:
            data["is_hidden"] = "true"

        if named_id is not None:
            data["named_id"] = named_id

        response = requests.post(url, headers=headers, data=json.dumps(data))

        if response.status_code == 200:
            r = response.json()
            return r

        raise Exception(response.text)

    def append_risk_score_doc_block(
        self, doc_id: int, query: str, named_id: str = None, is_hidden: bool = False
    ) -> dict:
        """
        Add a risk score block to a document. This risk score only evaluates based on what's in the document already; it does not query the fact base.

        Args:
            doc_id: the ID of the document to append the text block to
            query: the query to use for the AI block
            named_id: the named ID of the AI block
            is_hidden: whether the block is hidden in public documents
        Returns:
            JSON dict with document information
        """
        url = self.base_url + "api_doc_append_risk_score_doc_block"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {
            "doc_id": doc_id,
            "query": query,
        }

        if is_hidden:
            data["is_hidden"] = "true"

        if named_id is not None:
            data["named_id"] = named_id

        response = requests.post(url, headers=headers, data=json.dumps(data))

        if response.status_code == 200:
            r = response.json()
            return r

        raise Exception(response.text)

    def append_askai_fts_cot_block(
        self,
        doc_id: int,
        query: str,
        fts_term: str,
        named_id: str = None,
        is_hidden: bool = False,
    ) -> dict:
        """
        Add an AI (FTS) block to a document.

        Args:
            doc_id: the ID of the document to append the text block to
            query: the query to use for the AI block
            fts_term: the full text search (FTS) terms/query to use for the AI block
            named_id: the named ID of the AI block
            is_hidden: whether the block is hidden in public documents
        Returns:
            JSON dict with document information
        """

        url = self.base_url + "api_doc_append_askai_fts_cot_block"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {
            "doc_id": doc_id,
            "query": query,
            "arg": fts_term,
        }

        if is_hidden:
            data["is_hidden"] = "true"

        if named_id is not None:
            data["named_id"] = named_id

        response = requests.post(url, headers=headers, data=json.dumps(data))

        if response.status_code == 200:
            r = response.json()
            return r

        raise Exception(response.text)

    def append_askai_cot_block(
        self, doc_id: int, query: str, named_id: str = None, is_hidden: bool = False
    ) -> dict:
        """
        Add an AI (COT) block to a document.

        Args:
            doc_id: the ID of the document to append the text block to
            query: the query to use for the AI block
            named_id: the named ID of the AI block
            is_hidden: whether the block is hidden in public documents
        Returns:
            JSON dict with document information
        """

        url = self.base_url + "api_doc_append_askai_cot_block"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {
            "doc_id": doc_id,
            "query": query,
        }

        if is_hidden:
            data["is_hidden"] = "true"

        if named_id is not None:
            data["named_id"] = named_id

        response = requests.post(url, headers=headers, data=json.dumps(data))

        if response.status_code == 200:
            r = response.json()
            return r

        raise Exception(response.text)

    def append_askai_cot_pubdate_block(
        self, doc_id: int, query: str, named_id: str = None, is_hidden: bool = False
    ) -> dict:
        """
        Add an AI (COT) block to a document.

        Args:
            doc_id: the ID of the document to append the text block to
            query: the query to use for the AI block
            named_id: the named ID of the AI block
            is_hidden: whether the block is hidden in public documents
        Returns:
            JSON dict with document information
        """

        url = self.base_url + "api_doc_append_askai_cot_pubdate_block"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {
            "doc_id": doc_id,
            "query": query,
        }

        if is_hidden:
            data["is_hidden"] = "true"

        if named_id is not None:
            data["named_id"] = named_id

        response = requests.post(url, headers=headers, data=json.dumps(data))

        if response.status_code == 200:
            r = response.json()
            return r

        raise Exception(response.text)

    def append_askai_block(
        self, doc_id: int, query: str, named_id: str = None, is_hidden: bool = False
    ) -> dict:
        """
        Add an AI block to a document.

        Args:
            doc_id: the ID of the document to append the text block to
            query: the query to use for the AI block
            is_hidden: whether the block is hidden in public documents
        Returns:
            JSON dict with document information
        """

        url = self.base_url + "api_doc_append_askai_block"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {
            "doc_id": doc_id,
            "query": query,
        }

        if is_hidden:
            data["is_hidden"] = "true"

        if named_id is not None:
            data["named_id"] = named_id

        response = requests.post(url, headers=headers, data=json.dumps(data))

        if response.status_code == 200:
            r = response.json()
            return r

        raise Exception(response.text)

    def append_askai_block_fact_range(
        self, doc_id: int, query: str, named_id: str = None, is_hidden: bool = False, num_results=10, num_before=7, num_after=7
    ) -> dict:
        """
        Add an AI 'fact range' block to a document. This uses an experimental form of RAG that should reduce hallucinations but also reduces the # of sources that can be queried.

        Args:
            doc_id: the ID of the document to append the text block to
            query: the query to use for the AI block
            is_hidden: whether the block is hidden in public documents
            num_results: the number of results (sources) to return when doing a deep dive
            num_before: the number of facts to include prior to the core fact (with chunking)
            num_after: the number of facts to include after the core fact (with chunking)
        Returns:
            JSON dict with document information
        """

        url = self.base_url + "api_doc_append_askai_fact_range_block"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {
            "doc_id": doc_id,
            "query": query,
            "num_results": num_results,
            "num_before": num_before,
            "num_after": num_after
        }

        if is_hidden:
            data["is_hidden"] = "true"

        if named_id is not None:
            data["named_id"] = named_id

        response = requests.post(url, headers=headers, data=json.dumps(data))

        if response.status_code == 200:
            r = response.json()
            return r

        raise Exception(response.text)

    def create_automation_factbase(
        self, short_code, job_type, arg_string=None, args=None
    ):
        """
        Create an automation for a fact base.

        Args:
            short_code: the short code for the factbase
            job_type: the type of job to run
            arg_string: a string of arguments
            args: a dictionary of arguments if the job requires more than one

        Returns:
            dict: the automation object
        """

        url = self.base_url + "create_automation_for_factbase" + "/" + short_code
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {
            "job_type": job_type,
        }
        if arg_string != None:
            data["arg_string"] = arg_string
        if "args" != None:
            data["args"] = args
        response = requests.post(url, headers=headers, data=json.dumps(data))

        j = response.json()
        return j["info"]

    def create_automation_document(self, doc_id, job_type, arg_string=None, args=None):
        """
        Create an automation for a document.

        Args:
            doc_id: the document ID
            job_type: the type of job to run
            arg_string: a string of arguments
            args: a dictionary of arguments if the job requires more than one

        Returns:
            dict: the automation object
        """

        url = self.base_url + "create_automation_for_documente" + "/" + str(doc_id)
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {
            "job_type": job_type,
        }
        if arg_string != None:
            data["arg_string"] = arg_string
        if "args" != None:
            data["args"] = args
        response = requests.post(url, headers=headers, data=json.dumps(data))

        j = response.json()
        return j["info"]

    def get_factbase_automations(self, factbase_short_code):
        """
        Get an array of all automations that can be run on a factbase.

        Args:
            factbase_short_code: the short code for the factbase

        Returns:
            list: an array of automations that can be run on the factbase
        """
        url = self.base_url + f"get_automations/{factbase_short_code}"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        response = requests.post(url, headers=headers)
        j = response.json()
        print(j)
        return j["automations"]

    def queue_automation(self, automation_id):
        """
        Queue an automation (i.e., request the Emerging Trajectories platform to run an automation).

        Args:
            automation_id: the ID of the automation to queue

        Returns:
            bool: True if successful, False otherwise
        """
        url = self.base_url + f"queue_automation/{automation_id}"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        response = requests.post(url, headers=headers)
        if response.status_code == 200:
            return True
        return False

    def get_statement(self, statement_id: int) -> Statement:
        """
        Returns a given statement from the platform. Includes title, description, deadline, and fill-in-the-blank.

        Args:
            statement_id: the ID of the statement to retrieve

        Returns:
            Statement: the statement from the platform
        """
        url = self.base_url + "get_statement" + "/" + str(statement_id)
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        response = requests.post(url, headers=headers)
        if response.status_code == 200:
            r_obj = response.json()
            s = Statement(r_obj["title"], r_obj["fill_in_the_blank"])
            s.id = int(r_obj["id"])
            s.description = r_obj["description"]
            s.deadline = dateparser.parse(r_obj["deadline"])
            s.created_at = dateparser.parse(r_obj["created_at"])
            s.updated_at = dateparser.parse(r_obj["updated_at"])
            s.created_by = r_obj["created_by"]
            return s
        else:
            raise Exception(response.text)

    def get_forecast(self, forecast_id: int) -> Forecast:
        """
        Returns a given forecast from the platform.

        Args:
            forecast_id: the ID of the statement to retrieve

        Returns:
            Forecast: the forecast from the platform
        """
        url = self.base_url + "get_forecast" + "/" + str(forecast_id)
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        response = requests.post(url, headers=headers)
        if response.status_code == 200:

            r_obj = response.json()

            f = Forecast(r_obj["title"], float(r_obj["value"]), r_obj["justification"])

            f.id = int(r_obj["forecast_id"])

            f.statement_id = int(r_obj["statement_id"])
            f.statement = self.get_statement(int(r_obj["statement_id"]))

            f.created_at = dateparser.parse(r_obj["created_at"])
            f.updated_at = dateparser.parse(r_obj["updated_at"])
            # f.created_by = r_obj["created_by"]
            f.prediction_agent = r_obj["prediction_agent"]

            f.additional_data = r_obj["additional_data"]

            if "prior_forecast" in r_obj:
                if r_obj["prior_forecast"] is not None:
                    f.prior_forecast = int(r_obj["prior_forecast"])
            f.is_human = bool(r_obj["is_human"])

            if "next_forecasts" in r_obj:
                if r_obj is not None:
                    f.next_forecasts = r_obj["next_forecasts"]

            return f
        else:
            raise Exception(response.text)

    def get_facts_from_factbase(
        self,
        fact_db_slug: str,
        fact_id: str = None,
        fact_ids: list = None,
        source_url: str = None,
    ) -> list[dict]:
        """
        Gets a list of facts from a fact base.

        Args:
            fact_db_slug: the slug of the fact database to get facts from.
            fact_id: the ID of the fact to retrieve.
            fact_ids: a list of fact IDs to retrieve.
            source_url: the URL of the source to retrieve facts from.

        Returns:
            list: a list of facts from the fact base.
        """

        api_url = self.base_url + "get_facts/" + fact_db_slug
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        j = {}
        if fact_id is not None:
            j["fact_id"] = fact_id
        if fact_ids is not None:
            j["fact_ids"] = fact_ids
        if source_url is not None:
            j["source_url"] = source_url

        response = requests.post(api_url, headers=headers, json=j)

        if response.status_code == 200 or response.status_code == 201:
            return response.json()["facts"]

        raise Exception(response.text)

    def add_facts_to_factbase(
        self, fact_db_slug: str, url: str, facts: list[str]
    ) -> bool:
        """
        Adds a list of facts to a factbase on the Emerging Trajectories website.

        Args:
            fact_db_slug: the slug of the fact database to add the fact to.
            url: the URL of the fact.
            facts: the facts to add (a list of strings).

        Reutnr:
            bool: True if successful, False otherwise.
        """

        api_url = self.base_url + "add_facts/" + fact_db_slug
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        j = {
            "facts": facts,
            "url": url,
        }
        response = requests.post(api_url, headers=headers, json=j)

        if response.status_code == 200 or response.status_code == 201:
            return True
        return False

    def add_fact_to_factbase(self, fact_db_slug: str, url: str, fact: str) -> bool:
        """
        Adds a fact to a factbase on the Emerging Trajectories website.

        Args:
            fact_db_slug: the slug of the fact database to add the fact to.
            url: the URL of the fact.
            fact: the fact to add.

        Reutnr:
            bool: True if successful, False otherwise.
        """
        api_url = self.base_url + "add_fact/" + fact_db_slug
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        j = {
            "fact": fact,
            "url": url,
        }
        response = requests.post(api_url, headers=headers, json=j)

        if response.status_code == 200 or response.status_code == 201:
            return True
        print(response)
        return False

    def add_content_to_factbase(
        self, fact_db_slug: str, url: str, content: str, topic: str
    ) -> bool:
        """
        Sends content to the Emerging Trajectories website and extract facts from it.

        Args:
            fact_db_slug: the slug of the fact database to add the content to.
            url: the URL of the content. Note: we do not actually crawl this, we assume the content passed is the right conent.
            content: the content to extract facts from.
            topic: the topic of the content.

        Returns:
            bool: True if successful, False otherwise.
        """

        api_url = self.base_url + "add_content_to_factbase/" + fact_db_slug
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        j = {
            "content": content,
            "url": url,
            "topic": topic,
        }
        response = requests.post(api_url, headers=headers, json=j)

        if response.status_code == 200 or response.status_code == 201:
            return True
        print(response)
        return False

    def add_url_to_factbase(
        self, fact_db_slug: str, url: Union[str, list[str]], topic: str = ""
    ) -> bool:
        """
        Sends URL to the Emerging Trajectories website, crawls the URL, and extracts facts from it.

        Args:
            fact_db_slug: the slug of the fact database to add the content to.
            url: the URL of the content, which we will crawl. Could also be an array.
            topic: the topic of the content.

        Returns:
            bool: True if successful, False otherwise.
        """

        api_url = self.base_url + "add_url_to_factbase/" + fact_db_slug
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        j = {
            "url": url,
            "topic": topic,
        }
        response = requests.post(api_url, headers=headers, json=j)

        if response.status_code == 200 or response.status_code == 201:
            return True
        print(response)
        return False



"""
This is a convenience file for tracking prompts. We'll likely remove this in the (near) future.
"""

system_prompt_question_continuous = """You are a research agent that is meant to answer questions about specific points and topics. The facts you reference in answering these questions should all be based on information we provide you. We will provide you knowledge base below, where each fact is preceded by an ID (e.g., F1, F2, etc.). All your answers should be absed on these facts ONLY.

For example, suppose we ask, 'Who is the President of the USA?' and have the following facts...

F1: The President of the USA is Joe Biden.
F2: The Vice President of the USA is Kamala Harris.

... your answers hould be something like this:

The President of th USA is Joe Biden [F1].

We will give you a list of facts for every question. You can reference those facts using square brackets and the fact ID, so [F123] for fact 123, or you can also reference earlier facts from the conversation chain. YOU CANNOT USE OTHER INFORMATION."""



"""
This is an experimental approach to tracking information regardless of source type. It will also power more than recent updates. Here's how it works...

1. All "Content Sources" (a new class type that obtains content) will send content directly to the Facts DB.
2. The "Facts DB" will then extract all relevant facts for a prediction or research theme. It will keep cache the original content, will track the sources, and will also input all the facts into a RAG database.
3. We can then query the DB for relevant facts on an ad hoc basis, rather than only for new content.

"""

import os
import json
import hashlib
import re

# Using JSONEncoder to be consistent with the Emerging Trajectories website and platform.
from django.core.serializers.json import DjangoJSONEncoder

from phasellm.llms import OpenAIGPTWrapper, ChatBot, ChatPrompt
from phasellm.agents import WebpageAgent, WebSearchAgent

from datetime import datetime, timedelta

from . import Client
from .crawlers import crawlerPlaywright
from .prompts import *
from .news import NewsAPIAgent, RSSAgent, FinancialTimesAgent

import chromadb
import chromadb.utils.embedding_functions as embedding_functions

# Number of search results to return from web searche (default value).
_DEFAULT_NUM_SEARCH_RESULTS = 10

facts_base_system_prompt = """You are a researcher tasked with helping forecast economic and social trends. The title of our research project is: {statement_title}.

The project description is as follows...
{statement_description}

We will provide you with content from reports and web pages that is meant to help with the above. We will ask you to review these documents, create a set of bullet points to inform your thinking. Rather than using bullet points, please list each as F1, F2, F3, etc... So that we can reference it.

The content we provided you contains source numbers in the format 'SOURCE: #'. When you extract facts, please include the citation in square brackets, with the #, like [#], but replace "#" with the actual Source # from the crawled content we are providing you.

For example, if you are referring to a fact that came under --- SOURCE: 3 ---, you would write something like: "Data is already trending to hotter temperatures [3]." Do not include the "#" in the brackets, just the number.

Thus, a bullet point would look like this:
F1: (information) [1]
F2: (information) [1]
F3: (information) [2]

... and so on, where F1, F2, F3, etc. are facts, and [1], [2] are the source documents you are extracting the facts from.
"""

facts_base_user_prompt = """Today's date is {the_date}. We will now provide you with all the content we've managed to collect. 

----------------------
{scraped_content}
----------------------

Please think step-by-step by (a) extracting critical bullet points from the above, and (b) share any insights you might have based on the facts.

The content we provided you contains source numbers in the format 'SOURCE: #'. When you extract facts, please include the citation in square brackets, with the #, like [#], but replace "#" with the actual Source # from the crawled content we are providing you.

For example, if you are referring to a fact that came under --- SOURCE: 3 ---, you would write something like: "Data is already trending to hotter temperatures [3]." Do not include the "#" in the brackets, just the actual number.

DO NOT PROVIDE A FORECAST, BUT SIMPLY STATE AND SHARE THE FACTS AND INSIGHTS YOU HAVE GATHERED.
"""

fact_system_prompt = """You are a researcher helping extract facts about {topic}, trends, and related observations. We will give you a piece of content scraped on the web. Please extract facts from this. Each fact should stand on its own, and can be several sentences long if need be. You can have as many facts as needed. For each fact, please start it as a new line with "---" as the bullet point. For example:

--- Fact 1... This is the fact.
--- Here is a second fact.
--- And a third fact.

Please do not include new lines between bullet points. Make sure you write your facts in ENGLISH. Translate any foreign language content/facts/observations into ENGLISH.

We will simply provide you with content and you will just provide facts."""


def uri_to_local(uri: str) -> str:
    """
    Convert a URI to a local file name. In this case, we typically will use an MD5 sum.

    Args:
        uri (str): The URI to convert.

    Returns:
        str: The MD5 sum of the URI.
    """
    uri_md5 = hashlib.md5(uri.encode("utf-8")).hexdigest()
    return uri_md5


# TODO If this works, it should be an agent with setllm() supported, etc.
# TODO Right now, we don't actually save sources. It's an important feature (track reliability, etc. too!) but we want to ensure the POC works well first.
class FactRAGFileCache:

    def __init__(
        self,
        folder_path: str,
        openai_api_key: str,
        cache_file: str = "cache.json",
        sources_file: str = "sources.json",
        facts_file: str = "facts.json",
        rag_db_folder="cdb",
        crawler=None,
    ) -> None:
        """
        This is a RAG-based fact database. We build a database of facts available in JSON and via RAG and use this as a basic search engine for information. We use ChromaDB to index all facts, but also maintain a list of facts, sources, etc. in a JSON file. Finally, we keep a cache of all content and assume URLs do not get updated; we'll change this process in the future.

        Args:
            folder_path (str): The folder where everything will be stored.
            openai_api_key (str): The OpenAI API key. Used for RAG embeddings.
            cache_file (str, optional): The name of the cache file. Defaults to "cache.json".
            sources_file (str, optional): The name of the sources file. Defaults to "sources.json".
            facts_file (str, optional): The name of the facts file. Defaults to "facts.json".
            rag_db_folder (str, optional): The folder where the ChromaDB database will be stored. Defaults to "cdb".
            crawler (optional): The crawler to use. Defaults to None, in which case a Playwright crawler will be used.
        """
        self.root_path = folder_path
        self.root_parsed = os.path.join(folder_path, "parsed")
        self.root_original = os.path.join(folder_path, "original")
        self.cache_file = os.path.join(folder_path, cache_file)
        self.sources_file = os.path.join(folder_path, sources_file)
        self.facts_file = os.path.join(folder_path, facts_file)
        self.rag_db_folder = os.path.join(folder_path, rag_db_folder)
        self.openai_api_key = openai_api_key

        # Use the same default crawler for all other agents.
        # TODO Eventually we'll want to have an array agents we use to get content.
        if crawler is None:
            self.crawler = crawlerPlaywright()
        else:
            self.crawler = crawler

        # Set up / load Chroma DB
        openai_ef = embedding_functions.OpenAIEmbeddingFunction(
            api_key=self.openai_api_key, model_name="text-embedding-3-small"
        )
        self.chromadb_client = chromadb.PersistentClient(path=self.rag_db_folder)
        self.facts_rag_collection = self.chromadb_client.get_or_create_collection(
            name="facts", embedding_function=openai_ef
        )

        # Set up / load cache
        self.cache = self.load_cache()

        # Set up / load facts dictionary
        # TODO Eventually, move this to a database or table or something.
        self.facts = self.load_facts()

        # Set up / load sources dictionary
        # TODO Eventually, move this to a database or table or something.
        self.sources = self.load_sources()

    def query_to_fact_list(
        self, query: str, n_results: int = 10, since_date: datetime = None
    ) -> dict:
        """
        Takes a query and finds the closest semantic matches to the query in the knowledge base.

        Args:
            query (str): The query to search for.
            n_results (int, optional): The number of results to return. Defaults to 10.
            since_date (datetime, optional): The date to search from. Defaults to None, in which case all dates are searched.

        Returns:
            dict: A list of the facts found, with the key being the fact ID and each fact having its source, add date, and content info.
        """

        r = []
        if since_date is None:
            r = self.facts_rag_collection.query(
                query_texts=[query], n_results=n_results
            )
        else:
            r = self.facts_rag_collection.query(
                query_texts=[query],
                n_results=n_results,
                where={"added_on_timestamp": {"$gt": since_date.timestamp()}},
            )

        facts = {}
        for item in r["ids"][0]:
            facts[item] = {
                "content": self.facts[item]["content"],
                "source": self.facts[item]["source"],
                "added": self.facts[item]["added"],
            }

        return facts

    def query_to_fact_content(
        self, query: str, n_results: int = 10, since_date=None, skip_separator=False
    ) -> str:
        """
        Takes a query and finds the closest semantic matches to the query in the knowledge base.

        Args:
            query (str): The query to search for.
            n_results (int, optional): The number of results to return. Defaults to 10.
            since_date ([type], optional): The date to search from. Defaults to None, in which case all dates are searched.
            skip_separator (bool, optional): Whether to prepend and append a note horizontal line and title to the string being returned. Defaults to False.

        Returns:
            str: The content of the facts found, along with the fact IDs.

        """

        facts = self.query_to_fact_list(query, n_results, since_date)

        if len(facts) == 0:
            return ""

        fact_content = ""
        if not skip_separator:
            fact_content = """--- START FACTS ---------------------------\n"""

        for key, fact in facts.items():
            fact_content += key + ": " + fact["content"] + "\n"

        if not skip_separator:
            fact_content += """--- END FACTS ---------------------------\n"""

        return fact_content

    def get_all_recent_facts(self, days: float = 1, skip_separator=False) -> str:
        """
        Returns a list of all facts and sources added in the last n days.

        Args:
            days (float, optional): The number of days to search back. Defaults to 1. Can be fractional as well.
            skip_separator (bool, optional): Whether to prepend and append a note horizontal line and title to the string being returned. Defaults to False.

        Returns:
            str: The content of the facts found, along with the fact IDs.
        """

        fact_content = ""

        if not skip_separator:
            fact_content = """--- START FACTS ---------------------------\n"""

        min_date_timestamp = (datetime.now() - timedelta(days=days)).timestamp()
        for key, fact in self.facts.items():
            if fact["added_timestamp"] > min_date_timestamp:
                fact_content += key + ": " + fact["content"] + "\n"

        if not skip_separator:
            fact_content += """--- END FACTS ---------------------------\n"""

        return fact_content

    def save_facts_and_sources(self) -> None:
        """
        Saves facts and sources to their respective files.
        """
        with open(self.facts_file, "w") as f:
            json.dump(self.facts, f, indent=4, cls=DjangoJSONEncoder)
        with open(self.sources_file, "w") as f:
            json.dump(self.sources, f, indent=4, cls=DjangoJSONEncoder)

    def add_fact(self, fact: str, url: str) -> bool:
        """
        Adds a fact to the knowledge base.

        Args:
            fact (str): The fact to add.
            url (str): The URL source of the fact.

        Returns:
            bool: True if the fact was added, False otherwise.
        """

        fact_id_start = self.facts_rag_collection.count() + 1

        added_now = datetime.now()
        added_now_timestamp = added_now.timestamp()

        self.facts_rag_collection.add(
            documents=[fact],
            ids=[f"f{fact_id_start}"],
            metadatas=[{"added_on_timestamp": added_now_timestamp}],
        )

        self.facts[f"f{fact_id_start}"] = {
            "added": added_now,
            "added_timestamp": added_now_timestamp,
            "source": url,
            "content": fact,
            "cid": f"f{fact_id_start}",
        }

        self.save_facts_and_sources()

        return True

    def facts_from_url(self, url: str, topic: str) -> None:
        """
        Given a URL, extract facts from it and save them to ChromaDB and the facts dictionary. Also returns the facts in an array, in case one wants to analyze new facts.

        Args:
            url (str): Location of the content.
            topic (str): a brief description of the research you are undertaking.
        """

        content = self.get(url)

        llm = OpenAIGPTWrapper(self.openai_api_key, model="gpt-4-turbo-preview")
        chatbot = ChatBot(llm)
        chatbot.messages = [{"role": "system", "content": fact_system_prompt}]

        prompt_template = ChatPrompt(
            [
                {"role": "system", "content": fact_system_prompt},
            ]
        )

        chatbot.messages = prompt_template.fill(topic=topic)

        response = chatbot.chat(content)

        lines = response.split("\n")

        for line in lines:
            if line[0:4] == "--- ":
                fact = line[4:]
                self.add_fact(fact, url)

        self.save_facts_and_sources()

    # This builds facts based on RSS feeds.
    def new_get_rss_links(self, rss_url, topic) -> None:
        """
        Crawls an RSS feed and its posts.

        Args:
            rss_url (str): The URL of the RSS feed.
            topic (str): a brief description of the research you are undertaking.
        """

        rss_agent = RSSAgent(rss_url, crawler=self.crawler)
        urls = rss_agent.get_news_as_list()

        for url in urls:
            if not self.in_cache(url):
                print("RSS RESULT: " + url)
                try:
                    self.facts_from_url(url, topic)
                except:
                    print("Error; failed to get content from " + url)

    # This builds facts based on news articles.
    def new_get_new_info_news(
        self,
        newsapi_api_key,
        topic,
        queries,
        top_headlines=False,
    ) -> None:
        """
        Uses the News API to find new information and extract facts from it.

        Args:
            newsapi_api_key (str): The News API key.
            topic (str): a brief description of the research you are undertaking.
            queries (list[str]): A list of queries to search for.
            top_headlines (bool, optional): Whether to search for top headlines. Defaults to False.
        """

        news_agent = NewsAPIAgent(
            newsapi_api_key, top_headlines=top_headlines, crawler=self.crawler
        )

        for q in queries:
            results = news_agent.get_news_as_list(q)
            for result in results["articles"]:
                url = result["url"]
                if not self.in_cache(url):
                    print("NEWS RESULT: " + url)
                    self.facts_from_url(url, topic)

    # POC for FT
    def get_ft_news(self, ft_user, ft_pass, topic) -> None:
        """
        Uses the Financial Times Agent to find new information and extract facts from it.

        Args:
            ft_user (str): The Financial Times username.
            ft_pass (str): The Financial Times password.
            topic (str): a brief description of the research you are undertaking.
        """

        fta = FinancialTimesAgent(ft_user, ft_pass)
        urls, html_content, text_content = fta.get_news()

        if len(urls) != len(text_content):
            raise ValueError("URLs and text content are not the same length.")

        for i in range(0, len(urls)):
            url = urls[i]
            content = text_content[i]

            if not self.in_cache(url):
                print("FT RESULT: " + url)
                self.force_content(url, content)
                self.facts_from_url(url, topic)

    # This builds facts based on all the google searches.
    def new_get_new_info_google(
        self,
        google_api_key,
        google_search_id,
        google_search_queries,
        topic,
    ) -> None:
        """
        Uses Google search to find new information and extract facts from it.

        Args:
            google_api_key (str): The Google API key.
            google_search_id (str): The Google search ID.
            google_search_queries (list[str]): A list of queries to search for.
            topic (str): a brief description of the research you are undertaking.
        """

        self.google_api_key = google_api_key
        self.google_search_id = google_search_id
        self.google_search_queries = google_search_queries

        webagent = WebSearchAgent(api_key=self.google_api_key)

        for google_search_query in self.google_search_queries:

            results = webagent.search_google(
                query=google_search_query,
                custom_search_engine_id=self.google_search_id,
                num=_DEFAULT_NUM_SEARCH_RESULTS,
            )

            for result in results:
                if not self.in_cache(result.url):
                    try:
                        print("SEARCH RESULT: " + result.url)
                        # page_content = self.get(result.url)
                        self.facts_from_url(result.url, topic)
                        # print(page_content)
                    except Exception as e:
                        print(f"Failed to get content from {result.url}\n{e}")

    def save_state(self) -> None:
        """
        Saves the in-memory changes to the knowledge base to the JSON cache file.
        """
        with open(self.cache_file, "w") as f:
            json.dump(self.cache, f, cls=DjangoJSONEncoder)

    def load_facts(self) -> dict:
        """
        Loads the facts from the facts file.
        """
        if not os.path.exists(self.facts_file):
            with open(self.facts_file, "w") as f:
                f.write("{}")

        with open(self.facts_file, "r") as f:
            self.facts = json.load(f)

        return self.facts

    def load_sources(self) -> dict:
        """
        Loads the sources from the sources file.
        """
        if not os.path.exists(self.sources_file):
            with open(self.sources_file, "w") as f:
                f.write("{}")

        with open(self.sources_file, "r") as f:
            self.sources = json.load(f)

        return self.sources

    def load_cache(self) -> None:
        """
        Loads the cache from the cache file, or creates the relevant files and folders if one does not exist.
        """

        if not os.path.exists(self.root_path):
            os.makedirs(self.root_path)

        if not os.path.exists(self.root_parsed):
            os.makedirs(self.root_parsed)

        if not os.path.exists(self.root_original):
            os.makedirs(self.root_original)

        if not os.path.exists(self.cache_file):
            with open(self.cache_file, "w") as f:
                f.write("{}")

        with open(self.cache_file, "r") as f:
            return json.load(f)

    def in_cache(self, uri: str) -> bool:
        """
        Checks if a URI is in the cache already.

        Args:
            uri (str): The URI to check.

        Returns:
            bool: True if the URI is in the cache, False otherwise.
        """
        if uri in self.cache:
            return True
        return False

    def update_cache(
        self, uri: str, obtained_on: datetime, last_accessed: datetime
    ) -> None:
        """
        Updates the cache file for a given URI, specifically when it was obtained and last accessed.

        Args:
            uri (str): The URI to update.
            obtained_on (datetime): The date and time when the content was obtained.
            last_accessed (datetime): The date and time when the content was last accessed.
        """
        uri_md5 = uri_to_local(uri)
        self.cache[uri] = {
            "obtained_on": obtained_on,
            "last_accessed": last_accessed,
            "accessed": 0,
            "uri_md5": uri_md5,
        }
        self.save_state()

    def log_access(self, uri: str) -> None:
        """
        Saves the last accessed time and updates the accessed tracker for a given URI.

        Args:
            uri (str): The URI to update.
        """
        self.cache[uri]["last_accessed"] = datetime.now()
        self.cache[uri]["accessed"] = 1
        self.save_state()

    def get_unaccessed_content(self) -> list[str]:
        """
        Returns a list of URIs that have not been accessed by the agent.

        Returns:
            list[str]: A list of URIs that have not been accessed by the agent.
        """
        unaccessed = []
        for uri in self.cache:
            if self.cache[uri]["accessed"] == 0:
                unaccessed.append(uri)
        return unaccessed

    def force_content(self, uri: str, content: str, check_exists: bool = True) -> bool:
        """
        Forces a specific URI to have specific content (both HTML and text content). Used to fill old links that we don't actually want to crawl.

        Args:
            uri (str): The URI to force content for.
            content (str): The content to force.
            check_exists (bool): checks if content has already been included in the cache before forcing the new content.

        Returns:
            bool: True if the content was forced, False otherwise.
        """

        # If the content already exists and we avoid overwrites, then we don't want to overwrite it.
        if check_exists and self.in_cache(uri):
            return False

        uri_md5 = uri_to_local(uri)
        with open(os.path.join(self.root_original, uri_md5), "w") as f:
            f.write(content)
        with open(os.path.join(self.root_parsed, uri_md5), "w") as f:
            f.write(content)

        self.update_cache(uri, datetime.now(), datetime.now())
        self.log_access(uri)

        return True

    def get(self, uri: str) -> str:
        """
        Returns the content for a given URI. If the content is not in the cache, it will be scraped and added to the cache.

        Args:
            uri (str): The URI to get the content for.

        Returns:
            str: The content for the given URI.
        """
        uri_md5 = uri_to_local(uri)
        if uri in self.cache:
            with open(os.path.join(self.root_parsed, uri_md5), "r") as f:
                return f.read()
        else:
            # scraper = WebpageAgent()

            # content_raw = scraper.scrape(uri, text_only=False, body_only=False)
            # with open(os.path.join(self.root_original, uri_md5), "w") as f:
            #    f.write(content_raw)

            # content_parsed = scraper.scrape(uri, text_only=True, body_only=True)
            # with open(os.path.join(self.root_parsed, uri_md5), "w") as f:
            #    f.write(content_parsed)

            try:
                content, text = self.crawler.get_content(uri)
            except Exception as e:
                print(f"Failed to get content from {uri}\n{e}")
                content = ""
                text = ""

            with open(os.path.join(self.root_original, uri_md5), "w") as f:
                f.write(content)
            with open(os.path.join(self.root_parsed, uri_md5), "w") as f:
                f.write(text)

            self.update_cache(uri, datetime.now(), datetime.now())

            return text

    def add_content(self, content: str, uri: str = None) -> None:
        """
        Adds content to cache.

        Args:
            content (str): The content to add to the cache.
            uri (str, optional): The URI to use for the content. Defaults to None, in which case an MD5 sum of the content will be used.
        """
        if uri is None:
            uri = hashlib.md5(content.encode("utf-8")).hexdigest()
        uri_md5 = uri_to_local(uri)
        with open(os.path.join(self.root_parsed, uri_md5), "w") as f:
            f.write(content)
        self.update_cache(uri, datetime.now(), datetime.now())

    def add_content_from_file(self, filepath: str, uri: str = None) -> None:
        """
        Adds content from a text file to the cache.

        Args:
            filepath (str): The path to the file to add to the cache.
            uri (str, optional): The URI to use for the content. Defaults to None, in which case an MD5 sum of the content will be used.
        """
        with open(filepath, "r") as f:
            content = f.read()
        self.add_content(content, uri)


class FactBot:

    def __init__(
        self,
        knowledge_db: FactRAGFileCache,
        openai_api_key: str = None,
        chatbot: ChatBot = None,
    ) -> None:
        """
        The FactBot is like a ChatBot but enables you to ask questions that reference an underlying RAG database (KnowledgeBaseFileCache), which then enables the chatbot to cite sourcable facts.

        Args:
            knowledge_db (FactRAGFileCache): The knowledge database to use.
            openai_api_key (str, optional): The OpenAI API key. Defaults to None.
            chatbot (ChatBot, optional): The PhaseLLB chatbot to use. Defaults to None, in which case an OpenAI chatbot is used (and the OpenAI API key must be provided).
        """
        if openai_api_key is None and chatbot is None:
            raise ValueError("One of openai_api_key or chatbot must be provided.")

        if chatbot is not None:
            self.chatbot = chatbot
        else:
            llm = OpenAIGPTWrapper(openai_api_key, model="gpt-4-turbo-preview")
            self.chatbot = ChatBot(llm)
            self.chatbot.messages = [
                {"role": "system", "content": system_prompt_question_continuous}
            ]

        self.knowledge_db = knowledge_db

    def ask(self, question: str, clean_sources: bool = True) -> str:
        """
        Ask a question to the FactBot. This will query the underlying knowledge database and use the returned facts to answer the question.

        Args:
            question (str): The question to ask.
            clean_sources (bool, optional): Whether to clean the sources in the response. Defaults to True; in this case, it will replace fact IDs with relevant source links at the end of the response.

        Returns:
            str: The response to the question.
        """
        message = self.knowledge_db.query_to_fact_content(question) + "\n\n" + question
        response = self.chatbot.chat(message)
        if clean_sources:
            return clean_fact_citations(self.knowledge_db, response)
        else:
            return response

    def source(self, fact_id: str) -> str:
        """
        Returns the URL source for a given fact ID.

        Args:
            fact_id (str): The fact ID to get the source for.

        Returns:
            str: The URL source for the given fact ID.
        """
        if fact_id in self.knowledge_db.facts:
            return self.knowledge_db.facts[fact_id]["source"]

        if fact_id.lower() in self.knowledge_db.facts:
            return self.knowledge_db.facts[fact_id.lower()]["source"]

        raise ValueError(
            f"Fact ID " + str(fact_id) + " not found in the knowledge database."
        )

    def clean_and_source_to_html(
        self, text_to_clean: str, start_count: int = 0
    ) -> list:
        """
        Returns a formatted response with sourced HTML. This is used for emergingtrajectories.com and acts as a base for anyone else wanting to build similar features.

        Args:
            text_to_clean: The text to clean/cite/source.
            start_count: The starting count for the sources.

        Returns:
            list: two strings -- the actual response in the first case, and the sources in the second case, and an integer representing the new source count.
        """

        pattern = r"\[f[\d\s\,f]+\]"
        new_text = ""
        sources_text = ""
        ref_ctr = start_count
        last_index = 0

        for match in re.finditer(pattern, text_to_clean, flags=re.IGNORECASE):

            if match.group(0).find(",") == -1:
                ref_ctr += 1
                ref = match.group(0)[1:-1].strip()
                ref = ref.lower()

                new_text += text_to_clean[last_index : match.start()]
                new_text += f"""<a class='source_link' target='_blank' href='{self.source(ref)}'>{ref_ctr}</a>"""

                # Save the source
                fact_text = self.knowledge_db.facts[ref]["content"]
                new_source_text = f"""<span class='fact_span'><b>{ref_ctr}:</b> {fact_text} <a href='{self.source(ref)}' target='_blank'>View Source</a></span>"""
                sources_text += new_source_text + "\n"

                last_index = match.end()
            else:
                refs = match.group(0)[1:-1].split(",")
                ref_arr = []
                ref_str = ""
                for ref in refs:
                    ref = ref.strip()
                    ref = ref.lower()
                    ref_ctr += 1
                    ref_arr.append(str(ref_ctr))

                    # Add the source to the text
                    new_text_source_num = f"""<a class='source_link' target='_blank' href='{self.source(ref)}'>{ref_ctr}</a>"""
                    ref_str += " " + new_text_source_num

                    # Save the source
                    fact_text = self.knowledge_db.facts[ref]["content"]
                    new_source_text = f"""<span class='fact_span'><b>{ref_ctr}:</b> ${fact_text} <a href='{self.source(ref)}' target='_blank'>View Source</a></span>"""
                    sources_text += new_source_text + "\n"

                new_text += text_to_clean[last_index : match.start()] + ref_str
                last_index = match.end()

        new_text += text_to_clean[last_index:]

        return new_text, sources_text, ref_ctr


def clean_fact_citations(knowledge_db: FactRAGFileCache, text_to_clean: str) -> str:
    """
    Converts fact IDs referenced in a piece of text to relevant source links, appending sources as end notes in the document/text.

    Args:
        knowledge_db (FactRAGFileCache): The knowledge database to use for fact lookups.
        text_to_clean (str): The text to clean.

    Returns:
        str: The cleaned text.
    """
    bot = FactBot(knowledge_db, knowledge_db.openai_api_key)
    pattern = r"\[f[\d\s\,f]+\]"
    new_text = ""
    ref_ctr = 0
    last_index = 0
    sources_list = ""
    for match in re.finditer(pattern, text_to_clean):
        if match.group(0).find(",") == -1:
            ref_ctr += 1
            ref = match.group(0)[1:-1].strip()
            new_text += text_to_clean[last_index : match.start()]
            new_text += f"[{ref_ctr}]"
            sources_list += f"{ref_ctr} :: " + bot.source(f"{ref}") + "\n"
            last_index = match.end()
        else:
            refs = match.group(0)[1:-1].split(",")
            ref_arr = []
            for ref in refs:
                ref = ref.strip()
                ref_ctr += 1
                ref_arr.append(str(ref_ctr))
                sources_list += f"{ref_ctr} :: " + bot.source(f"{ref}") + "\n"
            ref_str = "[" + ", ".join(ref_arr) + "]"
            new_text += text_to_clean[last_index : match.start()] + ref_str
            last_index = match.end()

    new_text += text_to_clean[last_index:]

    if ref_ctr == 0:
        return text_to_clean
    else:
        return new_text + "\n\nSources:\n" + sources_list



"""
This is a very simple set of utility function(s) for loading PDF content. In fact, it might be easier to just use PyPDF directly and avoid this altogether. In the future, we might create specialized functions and classes for doing "fancy" things with PDFs (e.g., OCR, tables, etc.) so have created this module as a way to keep this in mind.
"""

from pypdf import PdfReader
import requests
import io


def get_PDF_content_from_file_by_page(file_path: str) -> list:
    """
    Loads a PDF file and extracts the text into a list of strings, one for each page.

    Args:
        file_path (str): The path to the PDF file.

    Returns:
        list: A list of strings, one for each page.
    """
    reader = PdfReader(file_path)
    content = []
    for page in reader.pages:
        content.append(page.extract_text())
    return content


def get_PDF_content_from_url_by_page(url: str) -> list:
    """
    Loads a PDF file from a URL and extracts the text into a list of strings, one for each page.

    Args:
        url (str): The URL to the PDF file.

    Returns:
        list: A list of strings, one for each page.
    """
    response = requests.get(url=url, timeout=120)
    pdf_file = io.BytesIO(response.content)
    reader = PdfReader(pdf_file)
    content = []
    for page in reader.pages:
        content.append(page.extract_text())
    return content


def get_PDF_content_by_page_from_file(file_path: str) -> str:
    """
    Loads a PDF file and extracts the text into one big string.

    Args:
        file_path (str): The path to the PDF file.

    Returns:
        str: The text content of the PDF file.
    """
    reader = PdfReader(file_path)
    content = ""
    for page in reader.pages:
        content += page.extract_text() + "\n"
    return content


def get_PDF_content_by_page_from_url(url: str) -> str:
    """
    Loads a PDF file from a URL and extracts the text into one big string.

    Args:
        url (str): The URL to the PDF file.

    Returns:
        str: The text content of the PDF file.
    """
    response = requests.get(url=url, timeout=120)
    pdf_file = io.BytesIO(response.content)
    reader = PdfReader(pdf_file)
    content = ""
    for page in reader.pages:
        content += page.extract_text() + "\n"
    return content



"""
Solutions for finding, extracting, storing, and reviisting knowledge.
"""

"""
This is the first knowledge base, and is meant to be a POC, really.

All of our agents as of today (Feb 1) focus on web searches and website content. Today, we do a Google search and scrape the content from the top results. We repeat this process every time the agent runs.

An obvious next step would be to create some sort of a cache to see if we already scraped the page and included the content elsewhere.

This should also be able to do multiple searches *and* accept other URLs to scrape.

How would this one work?
1. Have a folder where things get cached.
2. Have a JSON file that tracks when a knowledge base was accessed, the source URL, etc.
"""

import os
import json
import hashlib

# Using JSONEncoder to be consistent with the Emerging Trajectories website and platform.
from django.core.serializers.json import DjangoJSONEncoder

from phasellm.agents import WebpageAgent

from datetime import datetime

from . import Client
from phasellm.llms import OpenAIGPTWrapper, ChatBot

"""
CACHE STRUCTURE IN JSON...

key: URI (URL or file name)
value: {
    "obtained_on": <date>; when the file was downloaded
    "last_accessed": <date>; when the file was last used by the agent
    "accessed": 0 if not accessed, 1 if accessed
    "uri_md5": the MD5 sum of the URI
}

"""


def statement_to_search_queries(
    statement_id: int, client: Client, openai_api_key: str, num_queries: int = 3
) -> list[str]:
    """
    Given a specific statement ID, this will return a list of queries you can put into a search engine to get useful information.

    Args:
        statement_id (int): The ID of the statement to get search queries for.
        client (Client): The Emerging Trajectories API client.
        openai_api_key (str): The OpenAI API key.
        num_queries (int, optional): The number of queries to return. Defaults to 3.

    Returns:
        list[str]: A list of search queries.

    """

    statement = client.get_statement(statement_id)
    # print(statement)

    llm = OpenAIGPTWrapper(openai_api_key, model="gpt-3.5-turbo")
    chatbot = ChatBot(llm)

    chatbot.messages = [
        {
            "role": "system",
            "content": f"""I am working on a research project about this topic:\n{statement['title']}\n\n{statement['description']}\n\nHere is more information about what I am trying to do:\n{statement['description']}""",
        },
        {
            "role": "user",
            "content": "Could you please provide me with up to {num_queries} search queries that I can input into a search engine to find info about this topic? Please do not qualify your response... Simply provide one search query per line and nothing else.",
        },
    ]

    response = chatbot.resend()
    lines = response.strip().split("\n")

    if len(lines) > num_queries:
        lines = lines[:num_queries]

    return lines


def uri_to_local(uri: str) -> str:
    """
    Convert a URI to a local file name. In this case, we typically will use an MD5 sum.

    Args:
        uri (str): The URI to convert.

    Returns:
        str: The MD5 sum of the URI.
    """
    uri_md5 = hashlib.md5(uri.encode("utf-8")).hexdigest()
    return uri_md5


class KnowledgeBaseFileCache:

    def __init__(self, folder_path: str, cache_file: str = "cache.json") -> None:
        """
        The KnowledgeBaseFileCache is a simple file-based cache for web content and local files. The cache stores the original HTML, PDF, or TXT content and tracks when (if ever) an agent actually accessed the content.

        Args:
            folder_path (str): The folder where the cache will be stored.
            cache_file (str, optional): The name of the cache file. Defaults to "cache.json".
        """
        self.root_path = folder_path
        self.root_parsed = os.path.join(folder_path, "parsed")
        self.root_original = os.path.join(folder_path, "original")
        self.cache_file = os.path.join(folder_path, cache_file)
        self.cache = self.load_cache()

    def save_state(self) -> None:
        """
        Saves the in-memory changes to the knowledge base to the JSON cache file.
        """
        with open(self.cache_file, "w") as f:
            json.dump(self.cache, f, cls=DjangoJSONEncoder)

    def load_cache(self) -> None:
        """
        Loads the cache from the cache file, or creates the relevant files and folders if one does not exist.
        """

        if not os.path.exists(self.root_path):
            os.makedirs(self.root_path)

        if not os.path.exists(self.root_parsed):
            os.makedirs(self.root_parsed)

        if not os.path.exists(self.root_original):
            os.makedirs(self.root_original)

        if not os.path.exists(self.cache_file):
            with open(self.cache_file, "w") as f:
                f.write("{}")

        with open(self.cache_file, "r") as f:
            return json.load(f)

    def in_cache(self, uri: str) -> bool:
        """
        Checks if a URI is in the cache already.

        Args:
            uri (str): The URI to check.

        Returns:
            bool: True if the URI is in the cache, False otherwise.
        """
        if uri in self.cache:
            return True
        return False

    def update_cache(
        self, uri: str, obtained_on: datetime, last_accessed: datetime
    ) -> None:
        """
        Updates the cache file for a given URI, specifically when it was obtained and last accessed.

        Args:
            uri (str): The URI to update.
            obtained_on (datetime): The date and time when the content was obtained.
            last_accessed (datetime): The date and time when the content was last accessed.
        """
        uri_md5 = uri_to_local(uri)
        self.cache[uri] = {
            "obtained_on": obtained_on,
            "last_accessed": last_accessed,
            "accessed": 0,
            "uri_md5": uri_md5,
        }
        self.save_state()

    def log_access(self, uri: str) -> None:
        """
        Saves the last accessed time and updates the accessed tracker for a given URI.

        Args:
            uri (str): The URI to update.
        """
        self.cache[uri]["last_accessed"] = datetime.now()
        self.cache[uri]["accessed"] = 1
        self.save_state()

    def get_unaccessed_content(self) -> list[str]:
        """
        Returns a list of URIs that have not been accessed by the agent.

        Returns:
            list[str]: A list of URIs that have not been accessed by the agent.
        """
        unaccessed = []
        for uri in self.cache:
            if self.cache[uri]["accessed"] == 0:
                unaccessed.append(uri)
        return unaccessed

    def get(self, uri: str) -> str:
        """
        Returns the content for a given URI. If the content is not in the cache, it will be scraped and added to the cache.

        Args:
            uri (str): The URI to get the content for.

        Returns:
            str: The content for the given URI.
        """
        uri_md5 = uri_to_local(uri)
        if uri in self.cache:
            with open(os.path.join(self.root_parsed, uri_md5), "r") as f:
                return f.read()
        else:
            scraper = WebpageAgent()

            content_raw = scraper.scrape(uri, text_only=False, body_only=False)
            with open(os.path.join(self.root_original, uri_md5), "w") as f:
                f.write(content_raw)

            content_parsed = scraper.scrape(uri, text_only=True, body_only=True)
            with open(os.path.join(self.root_parsed, uri_md5), "w") as f:
                f.write(content_parsed)

            self.update_cache(uri, datetime.now(), datetime.now())

            return content_parsed

    def add_content(self, content: str, uri: str = None) -> None:
        """
        Adds content to cache.

        Args:
            content (str): The content to add to the cache.
            uri (str, optional): The URI to use for the content. Defaults to None, in which case an MD5 sum of the content will be used.
        """
        if uri is None:
            uri = hashlib.md5(content.encode("utf-8")).hexdigest()
        uri_md5 = uri_to_local(uri)
        with open(os.path.join(self.root_parsed, uri_md5), "w") as f:
            f.write(content)
        self.update_cache(uri, datetime.now(), datetime.now())

    def add_content_from_file(self, filepath: str, uri: str = None) -> None:
        """
        Adds content from a text file to the cache.

        Args:
            filepath (str): The path to the file to add to the cache.
            uri (str, optional): The URI to use for the content. Defaults to None, in which case an MD5 sum of the content will be used.
        """
        with open(filepath, "r") as f:
            content = f.read()
        self.add_content(content, uri)
